# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Union

import gemmi
import numpy as np
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Descriptor

from easycrystallography.Symmetry.SymOp import SymmOp

SG_DETAILS = {
    'space_group_HM_name': {
        'name': 'hermann_mauguin',
        'description': 'Hermann-Mauguin symbols given in Table 4.3.2.1 of International Tables for Crystallography '
        'Vol. A (2002) or a Hermann-Mauguin symbol for a conventional or unconventional setting.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_space_group/',
        'value': 'P 1',
    },
    'setting': {
        'name': 'coordinate-code',
        'description': 'A qualifier taken from the enumeration list identifying which setting in International Tables '
        'for Crystallography Volume A (2002) (IT) is used.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_space_group/',
        'value': '\x00',
    },
    'symmetry_ops': {
        'name': 'symmetry-ops',
        'description': 'A list of symmetry operations, each given as a 4x4 matrix in the form of a `SymmOp` object.',
        'value': [SymmOp.from_xyz_string('x,y,z')],
    },
}

if TYPE_CHECKING:
    import numpy.typing as npt
    from easyscience.Utils.typing import iF

    from easycrystallography.Components.Site import S

    T = Union[S, npt.ArrayLike]


_D_REDIRECT = deepcopy(Descriptor._REDIRECT)
_D_REDIRECT['value'] = lambda obj: ';'.join([r.as_xyz_string() for r in obj.raw_value.tolist()])


class easyOp(Descriptor):
    _REDIRECT = _D_REDIRECT


class SpaceGroup(BaseObj):
    _space_group_HM_name: ClassVar[Descriptor]
    _setting: ClassVar[Descriptor]
    _symmetry_ops: ClassVar[Descriptor]

    _REDIRECT = {'symmetry_ops': lambda obj: None if obj._sg_data is not None else obj._symmetry_ops}

    def __init__(
        self,
        space_group_HM_name: Optional[Descriptor, str] = None,
        setting: Optional[Descriptor, str] = None,
        symmetry_ops: Optional[List[SymmOp]] = None,
        interface: Optional[iF] = None,
    ):
        """
        Generate a spacegroup object from it's Hermann-Mauguin symbol and setting. The setting can be a part of the
        Hermann-Mauguin symbol.

        :param space_group_HM_name: Hermann-Mauguin symbol
        :param setting: Optional setting for the space group
        :param interface: Interface to the calculator
        """

        super(SpaceGroup, self).__init__(
            'space_group',
            _space_group_HM_name=Descriptor(**SG_DETAILS['space_group_HM_name']),
            _setting=Descriptor(**SG_DETAILS['setting']),
            _symmetry_ops=easyOp(**SG_DETAILS['symmetry_ops']),
        )

        if space_group_HM_name:
            self._space_group_HM_name = space_group_HM_name
        if setting is not None:
            self._setting = setting

        kwargs = {
            'new_spacegroup': self._space_group_HM_name.raw_value,
            'new_setting': self._setting.raw_value,
            'operations_set': None,
            'set_internal': True,
        }
        if symmetry_ops is not None:
            kwargs['operations_set'] = symmetry_ops
        self.__on_change(**kwargs)
        self.interface = interface
        self._cell = None

    @classmethod
    def from_int_number(cls, int_number: int, hexagonal=True, interface: Optional[iF] = None):
        """
        Generate a spacegroup object from it's spacegroup number (1-231).
        :param int_number: spacegroup number
        :param hexagonal: Should a hexagonal setting be used?
        :param interface: Interface to the calculator
        """
        sg = gemmi.find_spacegroup_by_number(int(int_number))
        setting = None
        if int_number in [146, 148, 155, 160, 161, 166, 167]:
            if hexagonal:
                setting = 'H'
            else:
                setting = 'R'
        return cls(sg.hm, setting, interface=interface)

    @classmethod
    def from_symOps(cls, sym_ops: List[SymmOp], interface: Optional[iF] = None):
        """
        Create a space group from a list of easycrystallography symmetry operations.

        :param sym_ops: List of easycrystallography symmetry operations
        :param interface: Interface to the calculator
        """
        ops = [gemmi.Op(op.as_xyz_string()) for op in sym_ops]
        sg_data = gemmi.find_spacegroup_by_ops(gemmi.GroupOps(ops))
        return cls(sg_data.hm, sg_data.ext, interface=interface)

    @classmethod
    def from_symMatrices(
        cls,
        rotations: List[np.ndarray],
        translations: List[np.ndarray],
        interface: Optional[iF] = None,
    ):
        """
        Create a space group from a lists of rotations and translations. The number of rotations and translations must
        be the same.

        :param rotations: Array of rotation matrices [n*[3x3]]
        :param translations: Array of translations [n*[1x3]]
        :param interface: Interface to the calculator
        """
        ops = []
        for rot, tran in zip(rotations, translations):
            ops.append(SymmOp.from_rotation_and_translation(rot, tran))
        return cls.from_symOps(ops, interface=interface)

    @classmethod
    def from_generators(
        cls,
        rotations: List[np.ndarray],
        translations: List[np.ndarray],
        interface: Optional[iF] = None,
    ):
        """
        Create a space group from a lists of rotations and translations. Each translation is applied to each rotation.

        :param rotations: Array of rotation matrices [n*[3x3]]
        :param translations: Array of translations [m*[1x3]]
        :param interface: Interface to the calculator
        """
        rots = len(translations) * rotations
        trans = [np.array(d) for d in np.array(translations).repeat(len(rotations), axis=0).tolist()]
        return cls.from_symMatrices(rots, trans, interface=interface)

    @classmethod
    def from_gemmi_operations(
        cls,
        operations: Union[gemmi.GroupOps, List[gemmi.Op]],
        interface: Optional[iF] = None,
    ):
        """
        Generate a space group from a list of gemmi operations or a gemmi group of operations.

        :param operations: Operations which define a space group
        :param interface: Interface to the calculator
        """
        ops = []
        if isinstance(operations, gemmi.GroupOps):
            ops = operations
        else:
            for op in operations:
                if isinstance(op, gemmi.Op):
                    ops.append(op)
                else:
                    raise TypeError('Operations must be of type gemmi.Op')
            ops = gemmi.GroupOps(ops)
        sg_data = gemmi.find_spacegroup_by_ops(ops)
        return cls(sg_data.hm, sg_data.ext, interface=interface)

    @classmethod
    def from_xyz_string(cls, xyz_string: Union[str, List[str]], interface: Optional[iF] = None):
        """
        Create a space group from a string or list of strings of the form 'x,y,z'.

        :param xyz_string: String defining space group operators
        :param interface: Interface to the calculator
        """
        if isinstance(xyz_string, str):
            xyz_string = xyz_string.split(';')
        ops = []
        for xyz in xyz_string:
            ops.append(SymmOp.from_xyz_string(xyz))
        return cls.from_symOps(ops, interface)

    @staticmethod
    def find_settings_by_number(number):
        """
        Find the IT_coordinate_system_code by group's number.
        gemmi doesn't do it natively.
        """
        ext = []
        for item in gemmi.spacegroup_table():
            if item.number > number:
                break
            if item.number == number:
                st = ''
                # Cases where ext and qualifier are not empty
                if item.ext and item.ext != '\x00':
                    st += item.ext
                if item.qualifier:
                    st += item.qualifier
                # special cases of defaul settings, not explicitly defined in gemmi
                system = item.crystal_system_str()
                if system == 'orthorhombic' and not item.qualifier:
                    st += 'abc'
                elif (system == 'trigonal' or system == 'hexagonal') and not item.qualifier:
                    st += 'h'
                # failed, just assign "1" to triclinic/monoclinic/tetragonal
                if not st:
                    st = '1'
                ext.append(st)
        return ext

    def __on_change(
        self,
        new_spacegroup: Union[int, str],
        new_setting: Optional[str] = None,
        operations_set: Optional[List[SymmOp]] = None,
        set_internal: bool = True,
    ) -> Tuple[str, str, List[SymmOp]]:
        """
        Internal function to update the space group. This function is called when the space group is changed. It checks
        the form of the imputs, generates reference data, and updates the internal data (if requested).

        :param new_spacegroup: New space group number or name
        :param new_setting: New space group setting
        :param set_internal: Should internal objects be updated
        """
        setting = '\x00'
        if operations_set is None:
            if isinstance(new_spacegroup, str):
                if ':' in new_spacegroup:
                    new_spacegroup, new_setting = new_spacegroup.split(':')
                sg_data = gemmi.find_spacegroup_by_name(new_spacegroup)
                if sg_data is None:
                    try:
                        sg_data = gemmi.find_spacegroup_by_ops(gemmi.symops_from_hall(new_spacegroup))
                    except RuntimeError:
                        sg_data = None
            else:
                sg_data = gemmi.find_spacegroup_by_number(int(new_spacegroup))

            if sg_data is None:
                raise ValueError(f"Spacegroup '{new_spacegroup}' not found in database.")

            settings = self.find_settings_by_number(sg_data.number)
            hm_name = sg_data.hm
            reference = sg_data.ext

            if new_setting is None or new_setting == '' or new_setting == '\x00':
                if reference != '\x00':
                    setting = reference
            else:
                try:
                    new_setting = int(new_setting)
                except ValueError:
                    pass
                new_setting = str(new_setting)
                if new_setting != reference:
                    # modify the space group with the new setting
                    new_sg_data = gemmi.find_spacegroup_by_name(sg_data.hm + ':' + new_setting)
                    if new_sg_data is None and new_setting in settings:
                        # this can be because the setting is the "default" setting which gemmi treats as a blank
                        new_sg_data = gemmi.find_spacegroup_by_name(sg_data.hm + ':' + reference)
                    if new_sg_data is None:
                        raise ValueError(f"Spacegroup '{new_spacegroup}:{new_setting}' not found in database.")
                    sg_data = new_sg_data
                    setting = sg_data.ext
                else:
                    setting = new_setting
                if new_setting in settings:
                    setting = new_setting

            if operations_set is None:
                operations_set = sg_data.operations()
            operations = [
                SymmOp.from_rotation_and_translation(np.array(op.rot) / op.DEN, np.array(op.tran) / op.DEN)
                for op in operations_set
            ]
        else:
            sg_data = None
            hm_name = 'custom'
            if isinstance(operations_set, str):
                operations_set = [SymmOp.from_xyz_string(s) for s in operations_set.split(';')]
            operations = operations_set
        if set_internal:
            self._sg_data = sg_data
            self._space_group_HM_name.value = hm_name
            self._setting.value = setting
            self._symmetry_ops.value = operations
        return sg_data, setting, operations

    @property
    def is_custom(self) -> bool:
        return self._sg_data is None

    @property
    def setting(self) -> Optional[Descriptor]:
        """
        Space group setting. If the space group does not have a setting, this will be None.

        :return: Space group setting
        """
        setting_str = self._setting.raw_value
        if setting_str == '\x00':
            return None  # no setting
        return self._setting

    @setting.setter
    def setting(self, new_setting: str) -> NoReturn:
        """
        Set the space group setting.

        :param new_setting: Space group setting
        """
        _, setting, _ = self.__on_change(self._space_group_HM_name.raw_value, new_setting, set_internal=True)

    @property
    def it_coordinate_system_code(self) -> Optional[Descriptor]:
        """
        Space group setting. If the space group does not have a setting, this will be None.
        Equivalent to setting, defined to satisfy CIF Template

        :return: Space group setting
        """
        setting_str = self._setting.raw_value
        if setting_str == '\x00':
            return None  # no setting
        return self._setting

    @it_coordinate_system_code.setter
    def it_coordinate_system_code(self, new_setting: str) -> NoReturn:
        """
        Set the space group setting.
        Equivalent to setting, defined to satisfy CIF Template

        :param new_setting: Space group setting
        """
        _, setting, _ = self.__on_change(self._space_group_HM_name.raw_value, new_setting, set_internal=True)

    @property
    def setting_str(self) -> str:
        """
        Space group setting as a string. If the space group does not have a setting, this will be an empty string.

        :return: Space group setting as a string
        """
        if self.setting is None:
            return ''
        return self._setting.raw_value

    @property
    def space_group_HM_name(self) -> Descriptor:
        """
        Space group name as defined by a Hermann-Mauguin symbol

        :return: Space group name as easyCore Descriptor
        """
        return self._space_group_HM_name

    @space_group_HM_name.setter
    def space_group_HM_name(self, value: str) -> NoReturn:
        """
        Set the space group name as defined by a Hermann-Mauguin symbol

        :param value: Space group name as a string
        """
        self.__on_change(value, set_internal=True)

    @property
    def name_hm_alt(self) -> Descriptor:
        """
        Space group name as defined by a Hermann-Mauguin symbol
        Equivalent to space_group_HM_name, defined to satisfy CIF Template

        :return: Space group name as easyCore Descriptor
        """
        return self._space_group_HM_name

    @name_hm_alt.setter
    def name_hm_alt(self, value: str) -> NoReturn:
        """
        Set the space group name as defined by a Hermann-Mauguin symbol
        Equivalent to space_group_HM_name, defined to satisfy CIF Template

        :param value: Space group name as a string
        """
        self.__on_change(value, set_internal=True)

    @property
    def hermann_mauguin(self) -> str:
        """
        Space group name as defined by a Hermann-Mauguin symbol

        :return: Space group name as a string
        """
        return self._space_group_HM_name.raw_value

    @property
    def name_hall(self) -> str:
        """
        Hall symbol of the space group
        Equivalent to hall_symbol, defined to satisfy CIF Template

        :return: Hall symbol of the space group
        """
        hall = None
        if not self.is_custom:
            hall = self._sg_data.hall
        return hall

    @property
    def hall_symbol(self) -> str:
        """
        Hall symbol of the space group

        :return: Hall symbol of the space group
        """
        hall = None
        if not self.is_custom:
            hall = self._sg_data.hall
        return hall

    @property
    def int_number(self) -> int:
        """
        International number of the space group

        :return: International number of the space group
        """
        n = None
        if not self.is_custom:
            n = self._sg_data.number
        return n

    @int_number.setter
    def int_number(self, new_it_number: int) -> NoReturn:
        """
        Set the spacegroup by its international number

        :param new_it_number: International number of the new space group
        """
        self.__on_change(new_it_number, set_internal=True)

    @property
    def it_number(self) -> int:
        """
        International number of the space group
        Equivalent to int_number, defined to satisfy CIF Template

        :return: International number of the space group
        """
        n = None
        if not self.is_custom:
            n = self._sg_data.number
        return n

    @it_number.setter
    def it_number(self, new_it_number: int) -> NoReturn:
        """
        Set the spacegroup by its international number
        Equivalent to int_number, defined to satisfy CIF Template

        :param new_it_number: International number of the new space group
        """
        self.__on_change(new_it_number, set_internal=True)

    @property
    def crystal_system(self) -> str:
        """
        Which crystal system the space group belongs to

        :return: Crystal system of the space group
        """
        s = ''
        if not self.is_custom:
            s = self._sg_data.crystal_system_str()
        return s

    @property
    def symmetry_ops(self) -> List[SymmOp]:
        """
        List of symmetry operations of the space group

        :return: List of symmetry operations of the space group
        """
        return self._symmetry_ops.raw_value

    @symmetry_ops.setter
    def symmetry_ops(self, new_ops: List[SymmOp]) -> NoReturn:
        """
        Set the symmetry operations of the space group

        :param new_ops: List of new symmetry operations
        """
        self.__on_change(
            self._space_group_HM_name.raw_value,
            operations_set=new_ops,
            set_internal=True,
        )

    @property
    def symmetry_xyz(self) -> str:
        """
        Symmetry operations of the space group as a string

        :return: String of symmetry operations of the space group
        """
        return ';'.join([op.as_xyz_string() for op in self._symmetry_ops.raw_value])

    @property
    def is_reference_setting(self) -> bool:
        """
        Is the space group setting the reference setting?

        :return: Is the space group setting the reference setting?
        """
        r = True
        if not self.is_custom:
            r = self._sg_data.is_reference_setting()
        return r

    def symmetry_matrices(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Get the rotational and translational matrices of the space group

        :return: Rotation and translation matrices
        """
        ops = self.symmetry_ops
        return [op.rotation_matrix.copy() for op in ops], [op.translation_vector.copy() for op in ops]

    def get_orbit(self, point: T, tol: float = 1e-5) -> np.ndarray:
        """
        Returns the orbit for a point.

        :param point: Point to get the orbit for
        :param tol: Tolerance for the orbit
        :return: Orbits of the point
        """
        orbit = []
        if not hasattr(point, '__iter__'):
            point = point.fract_coords
        point = np.array(point)
        for o in self.symmetry_ops:
            pp = np.array(o.operate(point))
            if not in_array_list(orbit, pp, tol=tol):
                orbit.append(pp)
        return np.array(orbit)

    def get_site_multiplicity(self, site: T, tol=1e-5) -> int:
        """
        Get the multiplicity of a given site

        :param site: Site to get the multiplicity of
        :param tol: Tolerance for the orbit
        :return: Multiplicity of the site
        """
        if not hasattr(site, '__iter__'):
            site = site.fract_coords
        site = np.array(site)
        multiplicity = 1
        for o in self.symmetry_ops:
            new_site = o.operate(site)
            if np.isclose(new_site, site, atol=tol).all():
                multiplicity += 1
        return multiplicity

    def __repr__(self) -> str:
        out_str = "<Spacegroup: system: '{:s}', number: {}, H-M: '{:s}'".format(
            self.crystal_system, self.int_number, self.hermann_mauguin
        )
        if self.setting_str:
            out_str = "{:s} setting: '{:s}'".format(out_str, self.setting_str)
        return out_str + '>'


def in_array_list(array_list, a, tol=1e-5) -> bool:
    """
    Extremely efficient nd-array comparison using numpy's broadcasting. This
    function checks if a particular array a, is present in a list of arrays.
    It works for arrays of any size, e.g., even matrix searches.

    Args:
        array_list ([array]): A list of arrays to compare to.
        a (array): The test array for comparison.
        tol (float): The tolerance. Defaults to 1e-5. If 0, an exact match is
            done.

    Returns:
        (bool)
    """
    if len(array_list) == 0:
        return False
    axes = tuple(range(1, a.ndim + 1))
    if not tol:
        return np.any(np.all(np.equal(array_list, a[None, :]), axes))
    return np.any(np.sum(np.abs(array_list - a[None, :]), axes) < tol)



"""
Constraints example
===================
This  example shows the usages of the different constraints.
"""

from easyscience.fitting import Constraints
from easyscience.Objects.ObjectClasses import Parameter

p1 = Parameter('p1', 1)
constraint = Constraints.NumericConstraint(p1, '<', 5)
p1.user_constraints['c1'] = constraint

for value in range(4, 7):
    p1.value = value
    print(f'Set Value: {value}, Parameter Value: {p1}')

#%%
# To include embedded rST, use a line of >= 20 ``#``'s or ``#%%`` between your
# rST and your code. This separates your example
# into distinct text and code blocks. You can continue writing code below the
# embedded rST text block:



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from typing import List

from .groups import SpaceGroup
from .groups import _get_symm_data


def _make_SG_names() -> list:
    sg_list = []
    for ind in range(1, 231):
        for op in SpaceGroup.SYMM_OPS:
            if op['number'] == ind:
                s = op['hermann_mauguin_fmt']
                if ':' in s:
                    s = s.split(':')[0]
                sg_list.append(s)
                break
    return sg_list


SG_NAMES = _make_SG_names()


class SpacegroupInfo:
    @staticmethod
    def get_all_systems() -> List[str]:
        return [
            'triclinic',
            'monoclinic',
            'orthorhombic',
            'tetragonal',
            'trigonal',
            'hexagonal',
            'cubic',
        ]

    @staticmethod
    def get_ints_from_system(system: str) -> List[int]:
        if system == 'triclinic':
            return list(range(1, 3))
        if system == 'monoclinic':
            return list(range(3, 16))
        if system == 'orthorhombic':
            return list(range(16, 75))
        if system == 'tetragonal':
            return list(range(75, 143))
        if system == 'trigonal':
            return list(range(143, 168))
        if system == 'hexagonal':
            return list(range(168, 195))
        return list(range(195, 231))

    @staticmethod
    def get_system_from_int(int_number: int):
        if int_number <= 2:
            return 'triclinic'
        if int_number <= 15:
            return 'monoclinic'
        if int_number <= 74:
            return 'orthorhombic'
        if int_number <= 142:
            return 'tetragonal'
        if int_number <= 167:
            return 'trigonal'
        if int_number <= 194:
            return 'hexagonal'
        return 'cubic'

    @staticmethod
    def get_symbol_from_int_number(int_number: int):
        return SG_NAMES[int_number - 1]

    @staticmethod
    def get_compatible_HM_from_int(int_number: int):
        return [sop['hermann_mauguin_fmt'] for sop in SpaceGroup.SYMM_OPS if sop['number'] == int_number]

    @staticmethod
    def get_compatible_HM_from_name(name: str):
        enc: dict = _get_symm_data('space_group_encoding')
        opt = enc.get(name.replace(' ', ''), None)
        if opt is None:
            raise AttributeError
        return SpacegroupInfo.get_compatible_HM_from_int(opt['int_number'])

    @staticmethod
    def get_int_from_HM(HM_str: str):
        ints = [sop['number'] for sop in SpaceGroup.SYMM_OPS if sop['hermann_mauguin'] == HM_str.replace(' ', '')]
        if not ints:
            raise AttributeError
        return ints[0]



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from typing import List
from typing import Tuple
from typing import Union

import numpy as np

from easycrystallography.Symmetry.SymOp import SymmOp


class Bonding:
    def __init__(
        self,
        dl: np.ndarray,
        atom1: np.ndarray,
        atom2: np.ndarray,
        idx: np.ndarray,
        n_sym: int,
    ):
        self.dl = dl
        self.atom1 = atom1
        self.atom2 = atom2
        self.idx = idx
        self.nSym = n_sym


def generate_bonds(
    phase_obj,
    force_no_sym: bool = False,
    max_distance: float = 8,
    tol: float = 1e-5,
    tol_dist: float = 1e-3,
    d_min: float = 0.5,
    max_sym: int = None,
    magnetic_only: bool = False,
) -> Bonding:
    """
    `generate_bonds` generates all bonds up to a certain length between magnetic atoms. It also groups bonds based
    either on crystal symmetry or bond length (with `tol_dist`  tolerance) is space group is not defined.
    Sorting bonds based on length can be forced by setting the `forceNoSym` parameter to true.

    :param phase_obj: easyCore phase object for which you want the bonds to be found
    :type phase_obj: Phase
    :param force_no_sym: Do not try to work out symmetry equivalent bonds.
    :type force_no_sym: bool
    :param max_distance: Maximum distance between atoms for which you want to search for a bond
    :type max_distance: float
    :param tol: The tolerance for which we consider unit cells.
    :type tol: float
    :param tol_dist: Tolerance of distance, within two bonds are considered equivalent
    :type tol_dist: float
    :param d_min: The minimum bond length between atoms
    :type d_min: float
    :param max_sym: Return only the 0->`max_sym` symmetry unique bonds.
    :type max_sym: int
    :param magnetic_only: Search for bonds which are between magnetic atoms only.
    :type magnetic_only: bool
    :return: Structure containing bond information
    :rtype: Bonding
    """

    max_distance = max_distance + tol

    # Calculate the height of the parallelepiped of a unit cell
    h_max1 = 1 / np.sqrt(np.sum(np.linalg.inv(phase_obj.cell.matrix.T) ** 2, axis=0))
    # calculate the distance of the[1 1 1] point from the origin
    h_max2 = np.abs(np.sum(phase_obj.cell.matrix, axis=0))

    # calculate the closest point to the origin
    h_max = np.min(np.array([h_max1, h_max2]), axis=0)

    # gives the number of extra unit cells along all 3 axes that are necessary to cover the minimum bond distance
    n_c = np.ceil(max_distance / h_max)

    all_atoms = phase_obj.get_orbits(magnetic_only=magnetic_only)
    all_atoms_r = np.vstack([np.array(all_atoms[key]) for key in all_atoms.keys()])
    n_atoms = all_atoms_r.shape[0]

    c_dim = np.array([n_c[0] + 1, 2 * n_c[1] + 1, 2 * n_c[2] + 1])

    # generate all cell translations
    c_tr1, c_tr2, c_tr3 = np.mgrid[0 : n_c[0] + 1, -n_c[1] : n_c[1] + 1, -n_c[2] : n_c[2] + 1]
    # cell origin translations: Na x Nb x Nc x 1 x 1 x3
    c_tr = np.stack([c_tr1, c_tr2, c_tr3], axis=3).reshape((*c_tr1.shape, 1, 1, 3))
    # remove unit cells that would produce duplicate bonds mark them with NaN (enough to do along a-axis values)
    c_tr[0, :, np.array(range(int(c_dim[2]))) < int(n_c[2]), 0, 0, 0] = np.nan
    c_tr[0, np.array(range(int(c_dim[1]))) < int(n_c[1]), int(n_c[2]), 0, 0, 0] = np.nan
    # % positions of atoms in the half 'cube' in l.u.
    # % Na x Nb x Nc x 1 x nMagAtom x 3
    # % atom2
    r1 = all_atoms_r.reshape((1, 1, 1, 1, -1, 3))
    r2 = all_atoms_r.reshape((1, 1, 1, -1, 1, 3))
    a_pos = r1 + c_tr
    # generate all distances from the atoms in the (0,0,0) cell in l.u.
    # r(atom2) - r(atom1)
    # Na x Nb x Nc x nMagAtom x nMagAtom x 3
    d_r = a_pos - r2
    # mark duplicate bonds within the (0,0,0) cell with nan
    r0 = d_r[0, int(n_c[1]), int(n_c[2]), :, :, :]
    d_r[0, int(n_c[1]), int(n_c[2]), :, :, :] = r0 * (
        1 + np.tril(np.nan * np.ones((n_atoms, n_atoms))).reshape((n_atoms, n_atoms, 1))
    )
    # calculate the absolute value of the distances in Angstrom
    d_ra = np.sqrt(np.sum(np.einsum('abcdei,ij->abcdej', d_r, phase_obj.cell.matrix) ** 2, axis=5))
    # reshape the numbers into a column list of bonds
    # 3 x Na x Nb x Nc x nMagAtom x nMagAtom
    dl = np.transpose(np.tile(c_tr, [1, 1, 1, n_atoms, n_atoms, 1]), axes=[5, 0, 1, 2, 3, 4])
    atoms1 = np.tile(
        np.arange(n_atoms).reshape((1, 1, 1, 1, -1, 1)),
        [1, *[int(d) for d in c_dim], 1, n_atoms],
    )
    atoms2 = np.tile(
        np.arange(n_atoms).reshape((1, 1, 1, 1, 1, -1)),
        [1, *[int(d) for d in c_dim], n_atoms, 1],
    )
    d_ra = d_ra.reshape((1, *d_ra.shape))
    # store everything in a single matrix
    # c_mat  = [dl(:,:);atom1(1,:);atom2(1,:);d_ra(1,:)];
    dl_ = dl.reshape((3, -1), order='F')
    atom1_ = atoms1.reshape((1, -1), order='F')
    atom2_ = atoms2.reshape((1, -1), order='F')
    d_ra_ = d_ra.reshape((1, -1), order='F')

    c_mat = np.vstack((dl_, atom1_, atom2_, d_ra_))
    # remove nan bonds
    c_mat = c_mat[:, ~np.isnan(c_mat[0, :])]
    # sort according to distance
    s_index = np.argsort(c_mat[5, :])
    c_mat = c_mat[:, s_index]
    # Apply cutoff
    c_index = c_mat[5, :] <= max_distance
    c_mat = c_mat[:, c_index]
    c_idx = np.cumsum(np.array(np.insert(np.diff(c_mat[5, :]) > tol_dist, 0, True), dtype=int)) - 1
    c_mat = np.vstack((c_mat, c_idx))

    if c_mat[5, 0] < d_min:
        raise ArithmeticError(f'Some atoms are too close (d_min={c_mat[5, 0]} < {d_min}), check your crystal structure!')

    d_ra = c_mat[5, :]
    c_mat = c_mat[[0, 1, 2, 3, 4, 6], :]
    basis_vector = phase_obj.cell.matrix
    sym_ops = phase_obj.spacegroup.symmetry_ops
    if not force_no_sym:
        n_mat = []
        if max_sym is None:
            max_sym = np.inf
        max_idx_sym = max(c_idx[c_idx <= max_sym])
        ii = 0
        idx = 0
        while ii <= max_idx_sym:
            sort_ms = c_mat[:, c_idx == ii]
            while sort_ms.shape[1] > 0:
                gen_c, un_c = bond(all_atoms_r, basis_vector, sort_ms[:, 0], sym_ops, tol)
                gen_c_all = np.hstack((gen_c, np.vstack((-gen_c[0:3, :], gen_c[[4, 3], :]))))
                # remove from sort_ms the identical couplings
                i_new, _ = isnew(gen_c_all, sort_ms[0:5, :], tol)
                sort_ms = sort_ms[:, i_new]
                # Remove identical couplings from the symmetry generated list
                gen_c = gen_c[:, un_c]
                if np.sum(~i_new) != np.sum(un_c):
                    raise ArithmeticError(f'Symmetry error! ii={ii}, idx={idx}. Try to change ' 'tol' ' parameter.')
                n_mat.append(np.vstack((gen_c, np.ones((1, gen_c.shape[1])) * idx)))
                idx += 1
            ii += 1
        n_mat = np.hstack(n_mat)
        # include the increase of bond index in case bonds are split due to symmetry inequivalency
        c_mat = c_mat[:, c_mat[5, :] > max_idx_sym]
        c_mat[5, :] = c_mat[5, :] + n_mat[5, -1] - max_idx_sym
        c_mat = np.hstack((n_mat, c_mat))

        # Save the value of maximum bond index that is generated by symmetry
        n_sym = n_mat[5, -1] + 1
    else:
        n_sym = 0

    return Bonding(
        c_mat[0:3, :].astype(int),
        c_mat[3, :].astype(int),
        c_mat[4, :].astype(int),
        c_mat[5, :].astype(int),
        n_sym,
    )


def bond(
    r: np.ndarray,
    bv: np.ndarray,
    single_bond: np.ndarray,
    sym_op: List[SymmOp],
    tol: float = 1e-5,
) -> Tuple[np.ndarray, Union[np.ndarray, np.ndarray]]:
    """
    generates all bonds that are symmetry equivalent to the given `bond`. The function uses the given space group
    operators and positions of magnetic atoms to return a list of equivalent bonds in a matrix. The function also checks
     the validity of the calculation by measuring the length of each equivalent bond using the given `bv` base and if
     the difference in length between equivalent bonds is larger than the tolerance throws a warning.

    :param r: Positions of the magnetic atoms in lattice units stored in a matrix
    :type r: np.ndarray
    :param bv: Basis vectors that define the lattice, used for checking the bond length of equivalent bonds
    :type bv: np.ndarray
    :param single_bond: Vector that contains the starting bond with elements of `[dl_a dl_b dl_c atom_1 atom_2]`,
    where `dl`
    is vector of lattice translation between the two atoms if they are not in the same unit cell in lattice units,
    `atom_1` and `atom_2` are indices of atoms in the list of positions stored in `r`.
    :type single_bond: np.ndarray
    :param sym_op: List of symmetry operations for the given spacegroup
    :type sym_op: list
    :param tol: Tolerance
    :type tol: float
    :return:
    :rtype:
    """
    tol_dist = 1e-5

    r1 = r[int(single_bond[3]), :].T
    r2 = r[int(single_bond[4]), :].T
    dl = single_bond[[0, 1, 2]]

    # Generate new atomic positions and translation vectors
    r1new = np.array([s.operate(r1) for s in sym_op]).T
    r2new = np.array([s.operate(r2) for s in sym_op]).T
    dlnew = np.array([s.apply_rotation_only(dl) for s in sym_op]).T - cfloor(r1new, tol) + cfloor(r2new, tol)

    # Modulo to get atoms in the first unit cell
    r1new = np.mod(r1new, 1)
    r2new = np.mod(r2new, 1)

    # Throw away generated couplings with wrong distance
    i_new, atom1 = isnewUC(r.T, r1new, tol_dist)
    if np.any(i_new):
        raise ArithmeticError('The generated positions for atom1 are wrong!')
    i_new, atom2 = isnewUC(r.T, r2new, tol_dist)
    if np.any(i_new):
        raise ArithmeticError('The generated positions for atom2 are wrong!')
    dist = np.sqrt(np.sum(np.dot(bv.T, (r.T[:, atom2] - r.T[:, atom1] + dlnew)) ** 2, axis=0))
    right_dist = np.abs(dist - dist[0]) < tol
    if not np.all(right_dist):
        # TODO This should raise a warning
        print('Symmetry generated couplings are dropped!')
    gen_cp = np.vstack((dlnew, atom1.reshape((1, -1)), atom2.reshape((1, -1))))
    gen_cp = gen_cp[:, right_dist]
    ugen_cp = uniqueB(gen_cp)

    return gen_cp, ugen_cp


def cfloor(r0: np.ndarray, tol: float) -> np.ndarray:
    """
    Floor for atomic positions

    floor(1-tol) == 1

    :param r0: Matrix of positions
    :type r0: np.ndarray
    :param tol: Tolerance
    :type tol: float
    :return: Matrix `r0` where values have been floored
    :rtype: np.ndarray
    """

    r = np.floor(r0)
    idx = np.abs(r0 - r) > 1 - tol
    r[idx] = r[idx] + 1
    return r


def isnewUC(
    matrix_a: np.ndarray, matrix_b: np.ndarray, tol: float
) -> Tuple[Union[np.ndarray, bool], Union[np.ndarray, int, float, complex]]:
    """
    Selects the new vectors from B within the first unit cell.

    Dimensions of A and B have to be [3 n_a] and [3 n_b] respectively. A vector in B is considered new,
    if d(mod(vA-vB,1))<tol.

    :param matrix_a: First matrix
    :type matrix_a: np.ndarray
    :param matrix_b: Second matrix
    :type matrix_b: np.ndarray
    :param tol: Tolerance
    :type tol: float
    :return: Logical numpy array saying of the vector is new and a numpy array with the index
    :rtype: Tuple[np.ndarray, np.ndarray]
    """

    n_a = matrix_a.shape[1]
    n_b = matrix_b.shape[1]

    C = matrix_a.shape[0]
    if C is not matrix_b.shape[0]:
        raise AttributeError('First dimension of A must be the same as that of B')

    aa = np.tile(matrix_a.reshape((*matrix_a.shape, 1)).transpose((1, 2, 0)), (1, n_b, 1))
    bb = np.tile(matrix_b.reshape((*matrix_b.shape, 1)).transpose((2, 1, 0)), (n_a, 1, 1))
    notequal = np.sum(cmod(aa - bb, tol) ** 2, axis=2) > tol

    is_new = np.all(notequal, axis=0)
    idx = np.arange(n_b)
    sym_idx = np.max(
        np.array(~notequal[:, idx[~is_new]], dtype=int) * np.arange(n_a).reshape((-1, 1)),
        axis=0,
    )

    return is_new, sym_idx


def isnew(
    matrix_a: np.ndarray, matrix_b: np.ndarray, tol: float
) -> Tuple[Union[np.ndarray, bool], Union[np.ndarray, int, float, complex]]:
    """
    Selects the new vectors from B within the first unit cell.

    Dimensions of A and B have to be [3 nA] and [3 nB] respectively. A vector in B is considered new,
    if d(mod(vA-vB,1))<tol.

    :param matrix_a: First matrix
    :type matrix_a: np.ndarray
    :param matrix_b: Second matrix
    :type matrix_b: np.ndarray
    :param tol: Tolerance
    :type tol: float
    :return: Logical numpy array saying of the vector is new and a numpy array with the index
    :rtype: Tuple[np.ndarray, np.ndarray]
    """

    if not matrix_a.shape[0] == matrix_b.shape[0]:
        raise AttributeError

    aa = np.tile(
        matrix_a.reshape((*matrix_a.shape, 1)).transpose((1, 2, 0)),
        (1, matrix_b.shape[1], 1),
    )
    bb = np.tile(
        matrix_b.reshape((*matrix_b.shape, 1)).transpose((2, 1, 0)),
        (matrix_a.shape[1], 1, 1),
    )
    notequal = np.sum(np.abs(aa - bb) ** 2, axis=2) > tol

    is_new = np.all(notequal, axis=0)
    idx = np.arange(matrix_b.shape[1])

    def sym_f(this_idx):
        return np.where(~notequal[:, this_idx])[0][0]

    sym_idx = np.array([sym_f(this_idx) for this_idx in idx[~is_new]])
    return is_new, sym_idx


def cmod(r: np.ndarray, tol: float) -> np.ndarray:
    """
    Modulus within tolerance

    :param r: Matrix to be operated on
    :type r: np.ndarray
    :param tol: Tolerance
    :type tol: float
    :return: `r` with modulus applied
    :rtype: np.ndarray
    """
    r = np.mod(r, 1)
    r[r > 1 - tol] = r[r > 1 - tol] - 1
    return r


def uniqueB(bond_matrix: np.ndarray) -> Union[np.ndarray, bool]:
    """
    Given an array of bonds, which ones are unique. Bonds has the shape [dl, atom1, atom2]

    :param bond_matrix: Array of bonds
    :type bond_matrix: np.ndarray
    :return: Logical array for if the corresponding bond is unique
    :rtype: np.ndarray
    """
    n_c = bond_matrix.shape[1]
    c1 = bond_matrix.reshape((*bond_matrix.shape, 1)).transpose((1, 2, 0))
    c2 = bond_matrix.reshape((*bond_matrix.shape, 1)).transpose((2, 1, 0))
    nc1 = np.vstack((-bond_matrix[0:3, :], bond_matrix[[4, 3], :])).reshape((5, -1, 1)).transpose((1, 2, 0))

    unique_b = np.all(
        np.triu(np.any(np.not_equal(c1, c2), axis=2) & np.any(np.not_equal(nc1, c2), axis=2))
        | np.tril(np.ones(n_c, dtype=bool)),
        axis=0,
    )
    return unique_b



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import numpy as np
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Descriptor
from easyscience.Objects.ObjectClasses import Parameter
from easyscience.Utils.classTools import addProp
from easyscience.Utils.classTools import removeProp

if TYPE_CHECKING:
    from easyscience.Utils.typing import iF

_ANIO_DETAILS = {
    'msp_type': {
        'description': 'A standard code used to describe the type of atomic displacement parameters used for the site.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 'Uani',
    },
    'Cani': {
        'description': 'Isotropic magnetic susceptibility parameter.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 0.0,
        'units': 'T^-1',
        'fixed': True,
    },
    'Ciso': {
        'description': 'Isotropic magnetic susceptibility parameter, or equivalent isotropic magnetic susceptibility '
        'parameter, C(equiv), in inverted teslas, calculated from anisotropic susceptibility '
        'components.',
        'value': 0.0,
        'max': np.inf,
        'units': 'T^-1',
        'fixed': True,
    },
}


class MSPBase(BaseObj):
    def __init__(self, *args, **kwargs):
        super(MSPBase, self).__init__(*args, **kwargs)

    @property
    def matrix(self) -> np.ndarray:
        matrix = np.zeros([3, 3])
        pars = self.get_parameters()
        if len(pars) == 1:
            np.fill_diagonal(matrix, pars[0].raw_value)
        elif len(pars) == 6:
            matrix[0, 0] = pars[0].raw_value
            matrix[0, 1] = pars[1].raw_value
            matrix[0, 2] = pars[2].raw_value
            matrix[1, 1] = pars[3].raw_value
            matrix[1, 2] = pars[4].raw_value
            matrix[2, 2] = pars[5].raw_value
        return matrix


class Cani(MSPBase):
    chi_11: ClassVar[Parameter]
    chi_12: ClassVar[Parameter]
    chi_13: ClassVar[Parameter]
    chi_22: ClassVar[Parameter]
    chi_23: ClassVar[Parameter]
    chi_33: ClassVar[Parameter]

    def __init__(
        self,
        chi_11: Optional[Union[Parameter, float]] = None,
        chi_12: Optional[Union[Parameter, float]] = None,
        chi_13: Optional[Union[Parameter, float]] = None,
        chi_22: Optional[Union[Parameter, float]] = None,
        chi_23: Optional[Union[Parameter, float]] = None,
        chi_33: Optional[Union[Parameter, float]] = None,
        msp_values: Optional[Type[MSPBase]] = None,
        interface=None,
    ):
        super(Cani, self).__init__(
            'Cani',
            chi_11=Parameter('chi_11', **_ANIO_DETAILS['Cani']),
            chi_12=Parameter('chi_12', **_ANIO_DETAILS['Cani']),
            chi_13=Parameter('chi_13', **_ANIO_DETAILS['Cani']),
            chi_22=Parameter('chi_22', **_ANIO_DETAILS['Cani']),
            chi_23=Parameter('chi_23', **_ANIO_DETAILS['Cani']),
            chi_33=Parameter('chi_33', **_ANIO_DETAILS['Cani']),
        )
        if chi_11 is not None:
            self.chi_11 = chi_11
        if chi_12 is not None:
            self.chi_12 = chi_12
        if chi_13 is not None:
            self.chi_13 = chi_13
        if chi_22 is not None:
            self.chi_22 = chi_22
        if chi_23 is not None:
            self.chi_23 = chi_23
        if chi_33 is not None:
            self.chi_33 = chi_33
        if msp_values is not None:
            self.chi_11 = msp_values.chi_11
            self.chi_12 = msp_values.chi_12
            self.chi_13 = msp_values.chi_13
            self.chi_22 = msp_values.chi_22
            self.chi_23 = msp_values.chi_23
            self.chi_33 = msp_values.chi_33
        self.interface = interface


class Ciso(MSPBase):
    chi: ClassVar[Parameter]

    def __init__(
        self,
        chi: Optional[Union[Parameter, float]] = None,
        msp_values: Optional[Type[MSPBase]] = None,
        interface: Optional[iF] = None,
    ):
        super(Ciso, self).__init__('Ciso', chi=Parameter('chi', **_ANIO_DETAILS['Ciso']))
        if chi is not None:
            self.chi = chi
        if msp_values is not None:
            self.chi = msp_values.chi
        self.interface = interface


_AVAILABLE_ISO_TYPES = {'Cani': Cani, 'Ciso': Ciso}


class MagneticSusceptibility(BaseObj):
    msp_type: ClassVar[Descriptor]
    msp_class: ClassVar[Type[MSPBase]]

    def __init__(self, msp_type: Union[Descriptor, str], interface: Optional = None, **kwargs):
        if isinstance(msp_type, str):
            msp_type = Descriptor('msp_type', msp_type)
        msp_class_name = msp_type.raw_value
        if msp_class_name in _AVAILABLE_ISO_TYPES.keys():
            msp_class = _AVAILABLE_ISO_TYPES[msp_class_name]
            if 'msp_class' in kwargs:
                # enable passing chi values directly to constructor
                kwargs['msp_values'] = kwargs['msp_class']
                _ = kwargs.pop('msp_class')
            msp = msp_class(**kwargs, interface=interface)
        else:
            raise AttributeError(f'{msp_class_name} is not a valid magnetic susceptibility type')
        super(MagneticSusceptibility, self).__init__('msp', msp_type=msp_type, msp_class=msp)
        for par in msp.get_parameters():
            addProp(
                self,
                par.name,
                fget=self.__a_getter(par.name),
                fset=self.__a_setter(par.name),
            )
        self.interface = interface

    def switch_type(self, msp_string: str, **kwargs):
        if msp_string in _AVAILABLE_ISO_TYPES.keys():
            msp_class = _AVAILABLE_ISO_TYPES[msp_string]
            if kwargs:
                msp_class: MSPBase = msp_class(interface=self.interface, **kwargs)
            else:
                msp_class: MSPBase = msp_class(interface=self.interface)
        else:
            raise AttributeError

        for par in self.msp_class.get_parameters():
            removeProp(self, par.name)
        self.msp_class = msp_class
        self.msp_type = msp_string
        for par in msp_class.get_parameters():
            addProp(
                self,
                par.name,
                fget=self.__a_getter(par.name),
                fset=self.__a_setter(par.name),
            )

    @property
    def available_types(self) -> List[str]:
        return [name for name in _AVAILABLE_ISO_TYPES.keys()]

    @staticmethod
    def __a_getter(key: str):
        def getter(obj):
            return obj.msp_class._kwargs[key]

        return getter

    @staticmethod
    def __a_setter(key):
        def setter(obj, value):
            obj.msp_class._kwargs[key].value = value

        return setter



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Dict
from typing import List
from typing import NoReturn
from typing import Tuple

from easycrystallography.Components.AtomicDisplacement import AtomicDisplacement as _AtomicDisplacement
from easycrystallography.Components.Site import Atoms as _Atoms
from easycrystallography.Components.Susceptibility import MagneticSusceptibility as _MagneticSusceptibility

from .template import CIF_Template
from .template import gemmi

if TYPE_CHECKING:
    from easyscience.Utils.typing import B


class AtomicDisplacement(CIF_Template):
    _CIF_SECTION_NAME: ClassVar[str] = '_atom_site'

    _CIF_ADP_ISO_CONVERSIONS = [
        ('label', '_label', '.label'),
        ('adp_type', '_adp_type', '.adp_type'),
        ('Biso', '_B_iso_or_equiv', '.B_iso_or_equiv'),
        ('Uiso', '_U_iso_or_equiv', '.U_iso_or_equiv'),
    ]
    _CIF_ADP_ANISO_CONVERSIONS = [
        ('label', '_aniso_label'),
        ('B_11', '_aniso_B_11'),
        ('B_12', '_aniso_B_12'),
        ('B_13', '_aniso_B_13'),
        ('B_22', '_aniso_B_22'),
        ('B_23', '_aniso_B_23'),
        ('B_33', '_aniso_B_33'),
        ('U_11', '_aniso_U_11'),
        ('U_12', '_aniso_U_12'),
        ('U_13', '_aniso_U_13'),
        ('U_22', '_aniso_U_22'),
        ('U_23', '_aniso_U_23'),
        ('U_33', '_aniso_U_33'),
    ]

    def __init__(self, reference_class=_AtomicDisplacement):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> Dict[str, B]:
        atom_dict = {}
        # ADP CHECKER
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_ADP_ISO_CONVERSIONS
        ]
        table = block.find(keys)
        if table.loop is None:
            # this means it might be the dictionary CIF
            keys = [
                self._CIF_SECTION_NAME + name[2] if 'label' in name[2] else '?' + self._CIF_SECTION_NAME + name[2]
                for name in self._CIF_ADP_ISO_CONVERSIONS
            ]
            table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['adp_type'] = row[1]
            else:
                if row.has(2):
                    kwargs['adp_type'] = 'Biso'
                elif row.has(3):
                    kwargs['adp_type'] = 'Uiso'
                else:
                    continue

            for i in range(2, 4):
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[kwargs['adp_type']] = V
                    if E:
                        errors[kwargs['adp_type']] = E
                    if F is not None and not F:
                        is_fixed[kwargs['adp_type']] = F
            obj = _AtomicDisplacement(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'adp': obj}

        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_ADP_ANISO_CONVERSIONS
        ]
        table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['adp_type'] = 'Bani'
                idx = 1
            elif row.has(7):
                idx = 7
                kwargs['adp_type'] = 'Uani'
            else:
                continue
            for i in range(idx, idx + 6):
                ll = self._CIF_ADP_ANISO_CONVERSIONS[i][0]
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[ll] = V
                    if E:
                        errors[ll] = E
                    if F is not None and not F:
                        is_fixed[ll] = F
            obj = _AtomicDisplacement(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'adp': obj}
        return atom_dict

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        # Add the additional anisotropic loops
        lines = []
        names = [self._CIF_ADP_ANISO_CONVERSIONS[0][1]]
        values_offset = 1
        # find the adp type first
        for atom in obj:
            if not hasattr(atom, 'adp'):
                continue
            if self.variable_to_string(atom.adp.adp_type) == 'Uani':
                values_offset = 7
            break
        for i in range(values_offset, values_offset + 6):
            names.append(self._CIF_ADP_ANISO_CONVERSIONS[i][1])

        for atom in obj:
            if not hasattr(atom, 'adp'):
                continue
            if not self.variable_to_string(atom.adp.adp_type) == 'Uani':  # non-anisotropic ADP on this atom
                continue
            line = [self.variable_to_string(atom.__getattribute__(self._CIF_ADP_ANISO_CONVERSIONS[0][0]))]
            # names.append(self._CIF_ADP_ANISO_CONVERSIONS[values_offset+i][1])
            for keys in self._CIF_ADP_ANISO_CONVERSIONS[values_offset : values_offset + 6]:
                key, cif_key = keys
                s = self.variable_to_string(atom.adp.__getattribute__(key))
                line.append(s)
            lines.append(line)

        # _, names = zip(*self._CIF_ADP_ANISO_CONVERSIONS[1])

        loop = block.init_loop(self._CIF_SECTION_NAME, names)
        for line in lines:
            loop.add_row(line)

    def iso_adp_block(self, obj: B, block: gemmi.cif.Block):
        # Add the isotropic objects
        labels = []
        objs = []

        if len(obj) == 0:
            return labels, objs
        adp0 = getattr(obj[0], 'adp', None)
        if adp0 is None:
            return labels, objs
        adp_type = self.variable_to_string(adp0.adp_type)
        labels = self._CIF_ADP_ISO_CONVERSIONS.copy()
        # should we have `Uani` in the line or only `Uiso`?
        if adp_type[0].lower() == 'u':
            del labels[2]
        else:
            del labels[3]
        ################################
        objs = [[getattr(_obj, 'adp')] for _obj in obj]
        del labels[0]
        return [[labels]] * len(objs), objs

    def from_cif_string(self, cif_string: str) -> List[Dict[str, B]]:
        if 'data_' not in cif_string:
            cif_string = 'data_temp\n' + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs


class MagneticSusceptibility(CIF_Template):
    _CIF_SECTION_NAME: ClassVar[str] = '_atom_site'
    _CIF_MSP_ANISO_CONVERSIONS = [
        ('label', '_susceptibility_label'),
        ('msp_type', '_susceptibility_chi_type'),
        ('chi_11', '_susceptibility_chi_11'),
        ('chi_12', '_susceptibility_chi_12'),
        ('chi_13', '_susceptibility_chi_13'),
        ('chi_22', '_susceptibility_chi_22'),
        ('chi_23', '_susceptibility_chi_23'),
        ('chi_33', '_susceptibility_chi_33'),
    ]

    def __init__(self, reference_class=_MagneticSusceptibility):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> Dict[str, B]:
        atom_dict = {}
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'label' in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_MSP_ANISO_CONVERSIONS
        ]
        table = block.find(keys)
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            if not row.has(0):
                continue
            if row.has(1):
                kwargs['msp_type'] = row[1]
            else:
                continue
            idx = 2
            for i in range(idx, idx + 6):
                ll = self._CIF_MSP_ANISO_CONVERSIONS[i][0]
                if row.has(i):
                    V, E, F = self.string_to_variable(row[i])
                    kwargs[ll] = V
                    if E:
                        errors[ll] = E
                    if F is not None and not F:
                        is_fixed[ll] = F
            obj = _MagneticSusceptibility(**kwargs)
            for error in errors.keys():
                setattr(getattr(obj, error), 'error', errors[error])
            for atr in is_fixed.keys():
                setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
            atom_dict[row[0]] = {'msp': obj}
        return atom_dict

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        # Then add the additional loops
        lines = []
        for atom in obj:
            if not hasattr(atom, 'msp'):
                continue
            line = [self.variable_to_string(atom.__getattribute__(self._CIF_MSP_ANISO_CONVERSIONS[0][0]))]
            for keys in self._CIF_MSP_ANISO_CONVERSIONS[1:]:
                key, cif_key = keys
                s = self.variable_to_string(atom.msp.__getattribute__(key))
                line.append(s)
            lines.append(line)
        _, names = zip(*self._CIF_MSP_ANISO_CONVERSIONS)
        loop = block.init_loop(self._CIF_SECTION_NAME, names)
        for line in lines:
            loop.add_row(line)

    def from_cif_string(self, cif_string: str) -> List[B]:
        if 'data_' not in cif_string:
            cif_string = 'data_temp\n' + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs


class Atoms(CIF_Template):
    _CIF_SECTION_NAME: ClassVar[str] = '_atom_site'
    _CIF_CONVERSIONS: ClassVar[List[Tuple[str, str]]] = [
        ('label', '_label', '.label'),
        ('specie', '_type_symbol', '.type_symbol'),
        ('fract_x', '_fract_x', '.fract_x'),
        ('fract_y', '_fract_y', '.fract_y'),
        ('fract_z', '_fract_z', '.fract_z'),
        ('occupancy', '_occupancy', '.occupancy'),
    ]

    def __init__(self, reference_class=_Atoms):
        super().__init__()
        self._CIF_CLASS = reference_class

    def _site_runner(self, block):
        keys = [
            self._CIF_SECTION_NAME + name[1] if 'occupancy' not in name[1] else '?' + self._CIF_SECTION_NAME + name[1]
            for name in self._CIF_CONVERSIONS
        ]
        table = (
            block.find(keys)
            if (table := block.find(keys)).loop is not None
            else block.find(
                [
                    self._CIF_SECTION_NAME + name[2] if 'occupancy' not in name[2] else '?' + self._CIF_SECTION_NAME + name[2]
                    for name in self._CIF_CONVERSIONS
                ]
            )
        )
        atom_dict = {}
        error_dict = {}
        fixed_dict = {}
        for row in table:
            kwargs = {}
            errors = {}
            is_fixed = {}
            for idx, item in enumerate(self._CIF_CONVERSIONS):
                ec_name, _, _ = item
                if row.has(idx):
                    V, E, F = self.string_to_variable(row[idx])
                    kwargs[ec_name] = V
                    if E:
                        errors[ec_name] = E
                    if F is not None and not F:
                        is_fixed[ec_name] = F
            atom_dict[kwargs['label']] = kwargs
            error_dict[kwargs['label']] = errors
            fixed_dict[kwargs['label']] = is_fixed

        # ADP CHECKER
        ADP_RUNNER = AtomicDisplacement()
        adp_dict = ADP_RUNNER.from_cif_block(block)
        for label, adp in adp_dict.items():
            if label in atom_dict:
                atom_dict[label].update(adp)

        # MSP Checker
        MSP_RUNNER = MagneticSusceptibility()
        msp_dict = MSP_RUNNER.from_cif_block(block)
        for label, msp in msp_dict.items():
            if label in atom_dict:
                atom_dict[label].update(msp)

        atoms = []
        for a in atom_dict.values():
            obj = self._CIF_CLASS._SITE_CLASS(**a)
            label = obj.label.raw_value
            if label in error_dict.keys():
                for atr in error_dict[label].keys():
                    setattr(getattr(obj, atr), 'error', error_dict[label][atr])
            if label in fixed_dict.keys():
                for atr in fixed_dict[label].keys():
                    setattr(getattr(obj, atr), 'fixed', fixed_dict[label][atr])
            atoms.append(obj)
        return atoms

    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        atoms = self._site_runner(block)
        return self._CIF_CLASS('from_cif', *atoms)

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        additional_keys = []
        additional_objs = []

        ADP_WRITER = AtomicDisplacement()
        if len(obj) > 0:
            if getattr(obj[0], 'adp', False):
                # additional keys for isotropic adp
                additional_keys, additional_objs = ADP_WRITER.iso_adp_block(obj, block)
        # Add main atom loop
        self._add_to_cif_block(obj, block, additional_keys, additional_objs)
        # another loop for anisotropic atomic displacements
        ADP_WRITER.add_to_cif_block(obj, block)
        # another loop for anisotropic magnetic susceptibilities
        MSP_WRITER = MagneticSusceptibility()
        MSP_WRITER.add_to_cif_block(obj, block)

    def _add_to_cif_block(self, obj: B, block: gemmi.cif.Block, additional_keys, additional_objs) -> NoReturn:
        # First add the main loop
        items = list(self._CIF_CONVERSIONS)
        names = [item[1] for item in items]
        lines = []
        # this loop adds atom name, symbol and coordinates as a line under the main "loop_"
        for idx1, atom in enumerate(obj):
            line = []
            for idx2, item in enumerate(items):
                if item[0] == 'specie':
                    s = str(atom.specie)
                else:
                    s = self.variable_to_string(atom.__getattribute__(item[0]))
                line.append(s)
            lines.append(line)

        # this loop adds ADP values to the end of the atom description lines
        for keys, objs, line in zip(additional_keys, additional_objs, lines):
            for idx, _obj in enumerate(objs):
                for key in keys[idx]:
                    k = key[0]
                    # Need to assure Uiso is present on ADP, else just assume 0.0
                    s = self.variable_to_string(_obj.__getattribute__(k)) if hasattr(_obj, k) else '0.0'
                    # for the cif representation, we only use [UB]iso
                    if 'ani' in s:
                        s = s.replace('ani', 'iso')
                    line.append(s)
                    if key[1] not in names:
                        names.append(key[1])

        # init_loop creates the main "loop_" section with '_atom_site_<names>' as the tags
        loop = block.init_loop(self._CIF_SECTION_NAME, names)
        for line in lines:
            loop.add_row(line)

    def from_cif_string(self, cif_string: str) -> List[B]:
        if 'data_' not in cif_string:
            cif_string = 'data_temp\n' + cif_string
        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

import textwrap
from abc import abstractmethod
from math import floor
from math import log10
from numbers import Number
from typing import TYPE_CHECKING
from typing import List
from typing import NoReturn
from typing import Optional

import gemmi
import numpy as np

if TYPE_CHECKING:
    from easyscience.Utils.typing import B
    from easyscience.Utils.typing import V

_MAX_LEN = 140
_MAX_LABEL_LEN = 130


class CIF_Template:
    """
    This class is an abstract class for a CIF template.
    """

    def __init__(self, decimal_places: int = 8):
        self.decimal_places = decimal_places

    @property
    def CLASS_READER(self):
        return self.from_cif_block

    @property
    def CLASS_WRITER(self):
        return self.add_to_cif_block

    @property
    def STRING_READER(self):
        return self.from_cif_string

    @property
    def STRING_WRITER(self):
        return self.to_cif_str

    @abstractmethod
    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        pass

    def to_cif_str(self, obj: B, block: Optional[gemmi.cif.Block] = None) -> str:
        """
        This method returns the CIF string.
        :return: The CIF string.
        """
        if block is None:
            block = gemmi.cif.Block('temp')
        start_str = block.as_string()
        self.add_to_cif_block(obj, block)
        final_str = block.as_string()
        return final_str[len(start_str) :]

    @abstractmethod
    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        pass

    @abstractmethod
    def from_cif_string(self, cif_string: str) -> List[B]:
        pass

    def variable_to_string(self, obj: V) -> str:
        value = obj.raw_value
        error = getattr(obj, 'error', None)
        if error is not None and np.isclose(error, 0.0):
            error = None
        decimal_places = self.decimal_places
        if isinstance(obj.raw_value, Number):
            if abs(value - int(value)) > 0.0:
                decimal_places = len(str(value).split('.')[1])
            initial_str = '{:.' + str(decimal_places) + 'f}'
            s = initial_str.format(round(value, decimal_places))
            if error is not None:
                xe_exp = int(floor(log10(error)))

                # uncertainty
                un_exp = xe_exp - decimal_places + 1
                un_int = round(error * 10 ** (-un_exp))

                # nominal value
                no_exp = un_exp
                no_int = round(value * 10 ** (-no_exp))

                # format - nom(unc)
                fmt = '%%.%df' % max(0, -no_exp)
                s = (fmt + '(%.0f)') % (
                    no_int * 10**no_exp,
                    un_int * 10 ** max(0, un_exp),
                )
        elif isinstance(value, str):
            s = '{:s}'.format(value)
        else:
            s = '{:s}'.format(str(value))
        if getattr(obj, 'fixed', None) is not None and not obj.fixed and error is None:
            s += '()'
        return CIF_Template._format_field(s)

    @staticmethod
    def string_to_variable(in_string: str):
        in_string = in_string.strip()
        if "'" in in_string:
            in_string = in_string.replace("'", '')
        if '"' in in_string:
            in_string = in_string.replace('"', '')
        fixed = None
        error = None
        tokens = in_string.split('(')
        try:
            value = float(tokens[0])
        except ValueError:
            value = tokens[0]
            return value, error, fixed
        if len(tokens) > 1:
            fixed = False
            if tokens[1][0] != ')':
                error = (10 ** -(len(f'{tokens[0]}'.split('.')[1]) - 1)) * int(tokens[1][:-1])
        return value, error, fixed

    @staticmethod
    def _format_field(v):
        v = v.__str__().strip()
        if len(v) > _MAX_LEN:
            return ';\n' + textwrap.fill(v, _MAX_LEN) + '\n;'
        # add quotes if necessary
        if v == '':
            return '""'
        if (' ' in v or v[0] == '_') and not (v[0] == "'" and v[-1] == "'") and not (v[0] == '"' and v[-1] == '"'):
            if "'" in v:
                q = '"'
            else:
                q = "'"
            v = q + v + q
        return v



# ruff: noqa
# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from typing import List
from typing import Optional
from typing import Union

from gemmi import cif

from .cif import *
from .template import AbstractStructureParser
from .template import AbstractStructureReader
from .template import AbstractStructureWriter


class CifFileReader(AbstractStructureReader):
    """
    Reads a structure from a CIF file.
    """

    def __init__(self, filename: str) -> None:
        """
        Initializes the CIF structure reader.

        :param filename: The filename of the CIF file.
        """
        self.file_path = filename
        self._handle = None
        self._document = cif.read(filename)

    def __exit__(self, *args, **kwargs):
        self._document = None

    def _block_finder(self, block_name):
        if block_name is None:
            block = self._document.sole_block()
        elif block_name == -1:
            block = self._document
        else:
            block = self._document.find_block(block_name)
        if block is None:
            raise AttributeError(f"Block '{block_name}' not found in {self.file_path}")
        return block

    def lattice(self, data_name: Optional[str] = None):
        block = self._block_finder(data_name)
        return self.read(block, Lattice().CLASS_READER)

    def atom_sites(self, data_name: Optional[str] = None):
        block = self._block_finder(data_name)
        return self.read(block, Atoms().CLASS_READER)

    def atom_site_aniso(self, data_name: Optional[str] = None):
        pass

    def atom_site_sus(self, data_name: Optional[str] = None):
        pass

    def symmetry(self, data_name: Optional[str] = None):
        block = self._block_finder(data_name)
        return self.read(block, SpaceGroup().CLASS_READER)

    def structure(self, data_name: Optional[str] = None, phase_class: Optional = None):
        block = self._block_finder(data_name)
        if phase_class is None:
            from easycrystallography.Structures.Phase import Phase as phase_class
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        kwargs = {'name': block.name}
        for key, value in components.items():
            kwargs[key] = self.read(block, value().CLASS_READER)
        return phase_class(**kwargs)

    def structures(self, phases_class: Optional = None, phase_class: Optional = None):
        if phase_class is None:
            from easycrystallography.Structures.Phase import Phase as phase_class
        if phases_class is None:
            from easycrystallography.Structures.Phase import Phases as phases_class
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        phases = []
        document = self._block_finder(-1)
        for block in document:
            kwargs = {'name': block.name}
            for key, value in components.items():
                kwargs[key] = self.read(block, value().CLASS_READER)
            phases.append(phase_class(**kwargs))
        return phases_class('from_cif', *phases)

    def read(self, in_str: str, reader_class):
        """
        Reads the CIF structure.

        :return: The CIF structure.
        """
        return reader_class(in_str)


class CifFileWriter(AbstractStructureWriter):
    """
    Writes a structure to a CIF file.
    """

    def __init__(self, filename: str, data_name: Optional[Union[str, List[str]]] = None, raw='w') -> None:
        """
        Initializes the CIF structure writer.

        :param filename: The filename of the CIF file.
        """
        self.filename = filename
        self.file = open(filename, raw)
        self.document = cif.Document()
        if isinstance(data_name, str):
            data_name = [data_name]
        self.data_name = data_name
        if self.data_name is None:
            self.data_name = []
        self.blocks = [self.document.add_new_block(name) for name in self.data_name]

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.file as w:
            w.write(self.document.as_string(cif.Style.Simple))

    def write(self, structure, writer_class, *args, **kwargs):
        """
        Writes the CIF structure.

        :param structure: The CIF structure.
        """
        return writer_class(structure, *args, **kwargs)

    def get_data_block(self, data_name: Optional[str] = None):
        if data_name is None:
            if len(self.document) == 0:
                data_name = 'data'
            else:
                data_name = self.document[0].name
        if data_name not in self.data_name:
            self.data_name.append(data_name)
            self.blocks.append(self.document.add_new_block(data_name))
        return self.document[self.data_name.index(data_name)]

    def lattice(self, obj, data_name: Optional[str] = None):
        block = self.get_data_block(data_name)
        return self.write(obj, Lattice().CLASS_WRITER, block)

    def atom_sites(self, obj, data_name: Optional[str] = None):
        block = self.get_data_block(data_name)
        return self.write(obj, Atoms().CLASS_WRITER, block)

    def atom_site_aniso(self, obj):
        pass

    def atom_site_sus(self, obj):
        pass

    def symmetry(self, obj, data_name: Optional[str] = None):
        block = self.get_data_block(data_name)
        return self.write(obj, SpaceGroup().CLASS_WRITER, block)

    def structure(self, obj, data_name: Optional[str] = None):
        if data_name is None:
            data_name = obj.name
        block = self.get_data_block(data_name)
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        for key, value in components.items():
            self.write(getattr(obj, key), value().CLASS_WRITER, block)

    def structures(self, objs):
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        for obj in objs:
            data_name = obj.name
            block = self.get_data_block(data_name)
            for key, value in components.items():
                self.write(getattr(obj, key), value().CLASS_WRITER, block)


class CifStringReader(AbstractStructureReader):
    """
    Reads a structure from a CIF String.
    """

    def __init__(self) -> None:
        """
        Initializes the CIF structure reader.

        """
        self._handle = None

    def __exit__(self, *args, **kwargs):
        pass

    def lattice(self, in_str):
        return self.read(in_str, Lattice().STRING_READER)

    def atom_sites(self, in_str):
        return self.read(in_str, Atoms().STRING_READER)

    def atom_site_aniso(self, in_str):
        pass

    def atom_site_sus(self, in_str):
        pass

    def symmetry(self, in_str):
        return self.read(in_str, SpaceGroup().STRING_READER)

    def structure(
        self,
        in_str: Optional[str] = None,
        block_name: Optional[str] = None,
        phase_class: Optional = None,
    ):
        if phase_class is None:
            from easycrystallography.Structures.Phase import Phase as phase_class
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        document = cif.read_string(in_str)
        data_names = [block.name for block in document]
        idx = 0
        if block_name is not None:
            if block_name in data_names:
                idx = data_names.index(block_name)
            else:
                raise ValueError('Block name not found')
        block = document[idx]
        kwargs = {'name': block.name}
        for key, value in components.items():
            kwargs[key] = self.read(in_str, value().STRING_READER)[idx]
        return phase_class(**kwargs)

    def structures(self, document, phase_class: Optional = None):
        from easycrystallography.Structures.Phase import Phases

        if phase_class is None:
            from easycrystallography.Structures.Phase import Phase as phase_class
        components = {'cell': Lattice, 'spacegroup': SpaceGroup, 'atoms': Atoms}
        phases = []
        document = cif.read_string(document)
        for block in document:
            kwargs = {'name': block.name}
            for key, value in components.items():
                kwargs[key] = self.read(block, value().CLASS_READER)
            phases.append(phase_class(**kwargs))
        return Phases('from_cif', *phases)

    def read(self, in_str: str, reader_class):
        """
        Reads the CIF structure.

        :return: The CIF structure.
        """
        return reader_class(in_str)


class CifStringWriter(AbstractStructureWriter):
    """
    Writes a structure to a CIF file.
    """

    def __init__(self) -> None:
        """
        Initializes the CIF structure writer.
        """
        self._handle = None

    def __exit__(self, *args, **kwargs):
        pass

    def lattice(self, obj) -> str:
        return self.write(obj, Lattice().STRING_WRITER)

    def atom_sites(self, obj):
        return self.write(obj, Atoms().STRING_WRITER)

    def atom_site_aniso(self, obj):
        pass

    def atom_site_sus(self, obj):
        pass

    def symmetry(self, obj) -> str:
        return self.write(obj, SpaceGroup().STRING_WRITER)

    def structure(self, obj, data_name: Optional[str] = None):
        d = cif.Document()
        if data_name is None:
            data_name = obj.name
        block = d.add_new_block(data_name)
        components = {'cell': Lattice, 'space_group': SpaceGroup, 'atoms': Atoms}
        for key, value in components.items():
            self.write(getattr(obj, key), value().CLASS_WRITER, block)
        return d.as_string(cif.Style.Simple)

    def structures(self, objs):
        components = {'cell': Lattice, 'space_group': SpaceGroup, 'atoms': Atoms}
        d = cif.Document()
        for obj in objs:
            data_name = obj.name
            block = d.add_new_block(data_name)
            for key, value in components.items():
                self.write(getattr(obj, key), value().CLASS_WRITER, block)
        return d.as_string(cif.Style.Simple)

    def write(self, structure, writer_class, *args, **kwargs) -> str:
        """
        Writes the CIF structure.

        :param structure: The CIF structure.
        """
        return writer_class(structure, *args, **kwargs)


class CifStringParser(AbstractStructureParser):
    def __init__(self):
        super(CifStringParser, self).__init__(CifStringReader, CifStringWriter)


class CifFileParser(AbstractStructureParser):
    def __init__(self):
        super(CifFileParser, self).__init__(CifFileReader, CifFileWriter)



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from .cif_parser import CifFileParser
from .cif_parser import CifStringParser


class Parsers:
    _parsers = {'cif_str': CifStringParser, 'cif': CifFileParser}

    def __init__(self, parser_choice: str, *args, **kwargs):
        self._setup_parser(parser_choice, *args, **kwargs)

    def _setup_parser(self, parser_choice, *args, **kwargs):
        try:
            p = self._parsers.get(parser_choice)
        except KeyError:
            raise KeyError(f"The parser '{parser_choice}' is not a known parser")
        self._parser = p(*args, **kwargs)

    def writer(self, *args, **kwargs):
        return self._parser.writer(*args, **kwargs)

    def reader(self, *args, **kwargs):
        return self._parser.reader(*args, **kwargs)

    @property
    def parsers(self):
        return list(self._parsers.keys())

    @property
    def parser(self):
        return self._parser

    def set_parser(self, new_parser, *args, **kwargs):
        self._setup_parser(new_parser, *args, **kwargs)



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

import re
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Union

import periodictable as pt
from easyscience.Objects.ObjectClasses import Descriptor
from easyscience.Utils.classTools import addProp
from easyscience.Utils.classTools import removeProp

_SPECIE_DETAILS = {
    'type_symbol': {
        'description': 'A code to identify the atom species occupying this site.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
    },
}

_REDIRECT = deepcopy(Descriptor._REDIRECT)
_REDIRECT['specie'] = lambda obj: obj._raw_data['str']


class Specie(Descriptor):
    _REDIRECT = _REDIRECT

    def __init__(self, specie: str = 'H', **kwargs):
        if 'value' in kwargs.keys():
            specie = kwargs.pop('value')

        self._raw_data: Dict[str, Union[str, int]] = {}
        self._props: Dict[str, Any] = {}

        self._reset_data()
        super(Specie, self).__init__('specie', self.__gen_data(specie), **_SPECIE_DETAILS['type_symbol'])

        # Monkey patch the unit and the value to take into account the new type situation
        self.__previous_set = self.__class__.value.fset

        addProp(
            self,
            'value',
            fget=self.__class__.value.fget,
            fset=lambda obj, val: self.__previous_set(obj, obj.__gen_data(val)),
            fdel=self.__class__.value.fdel,
        )

    def _reset_data(self):
        self._raw_data = dict.fromkeys(['str', 'observed', 'element', 'isotope', 'oxi_state', 'spin'])
        self._props = {}

    def __gen_data(self, value_str: str):
        s = re.search(r'([0-9.]*)([A-Z][a-z]*)([0-9.]*)([+\-]*)', value_str)
        # group(1) = Isotope
        # group(2) = Element
        # group(3) = Oxi state
        # group(4) = Oxi +/2

        if s is None:
            raise ValueError('Invalid specie string')
        element_str = s.group(2)
        element = getattr(pt, element_str, None)
        if element is None:
            raise ValueError(f'Element ({s.group(2)}) not found in periodictable')

        if self._props is not None:
            for k in self._props.keys():
                removeProp(self, k)

        self._reset_data()
        self._raw_data['str'] = value_str
        self._raw_data['element'] = element
        self._raw_data['observed'] = element

        props = {
            s: getattr(self._raw_data['observed'], s)
            for s in self._raw_data['observed'].__dir__()
            if not s.startswith('_') and hasattr(self._raw_data['observed'], s)
        }
        isotope_str = s.group(1)
        self._raw_data['isotope'] = None
        if isotope_str:
            self._raw_data['isotope'] = int(isotope_str)
            self._raw_data['observed'] = pt.core.Isotope(self._raw_data['observed'], self._raw_data['isotope'])
            props.update(
                {
                    s: getattr(self._raw_data['observed'], s)
                    for s in self._raw_data['observed'].__dir__()
                    if not s.startswith('_') and hasattr(self._raw_data['observed'], s)
                }
            )
        oxi_state_str = s.group(3)
        oxi_pm_str = s.group(4)
        self._raw_data['oxi_state'] = None
        if oxi_state_str:
            self._raw_data['oxi_state'] = int(oxi_pm_str + oxi_state_str)
            self._raw_data['observed'] = pt.core.Ion(self._raw_data['observed'], self._raw_data['oxi_state'])
            props.update(
                {
                    s: getattr(self._raw_data['observed'], s)
                    for s in self._raw_data['observed'].__dir__()
                    if not s.startswith('_') and hasattr(self._raw_data['observed'], s)
                }
            )
        props['common_name'] = props.pop('name')
        self._props = props
        for k, v in props.items():
            addProp(self, k, fget=self.__getter_periodic(k))
        rep = f"{self._raw_data['element']}"
        if self.is_ion:
            rep += f"{abs(self._raw_data['oxi_state'])}"
            if self._raw_data['oxi_state'] > 0:
                rep += '+'
            else:
                rep += '-'
        return rep

    def __repr__(self) -> str:
        rep = f"<{self.__class__.__name__} '{self.name}': "
        if self.is_isotope:
            rep += f"{self._raw_data['isotope']}"
        rep += f"{self._raw_data['element']}"
        if self.is_ion:
            if self._raw_data['oxi_state'] > 0:
                rep += f"{self._raw_data['oxi_state']}+"
            else:
                rep += f"{self._raw_data['oxi_state']}-"
        rep += '>'
        return rep

    def __str__(self) -> str:
        rep = ''
        if self.is_isotope:
            rep += f"{self._raw_data['isotope']}"
        rep += f"{self._raw_data['element']}"
        if self.is_ion:
            if self._raw_data['oxi_state'] > 0:
                rep += f"{self._raw_data['oxi_state']}+"
            else:
                rep += f"{str(self._raw_data['oxi_state'])[1:]}-"
        return rep

    @staticmethod
    def __getter_periodic(key: str):
        def getter(obj):
            return obj._props.get(key)

        return getter

    @property
    def is_ion(self):
        return self._raw_data['oxi_state'] is not None

    @property
    def is_isotope(self):
        return self._raw_data['isotope'] is not None



"""
Subclassing BaseObj - Simple Pendulum
=====================================
This  example shows how to subclass :class:`easyscience.Objects.Base.BaseObj` with parameters from
:class:`easyscience.Objects.Base.Parameter`. For this example a simple pendulum will be modeled.

.. math::
    y = A \sin (2 \pi f t + \phi )

Imports
*******

Firstly the necessary imports. Notice that we import numpy from easyscience. This is not done for any reason other than
saving time from multiple imports.
"""
import matplotlib.pyplot as plt
import numpy as np
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Parameter

#%%
# Subclassing
# ***********
# To include embedded rST, use a line of >= 20 ``#``'s or ``#%%`` between your
# rST and your code. This separates your example
# into distinct text and code blocks. You can continue writing code below the
# embedded rST text block:


class Pendulum(BaseObj):
    def __init__(self, A: Parameter, f: Parameter, p: Parameter):
        super(Pendulum, self).__init__('SimplePendulum', A=A, f=f, p=p)

    @classmethod
    def from_pars(cls, A: float = 1, f: float = 1, p: float = 0):
        A = Parameter('Amplitude', A)
        f = Parameter('Frequency', f)
        p = Parameter('Phase', p)
        return cls(A, f, p)

    def __call__(self, t):
        return self.A.raw_value * np.sin(2*np.pi*self.f.raw_value*t + self.p.raw_value)

    def plot(self, time, axis=None, **kwargs):
        if axis is None:
            axis = plt
        else:
            axis.set_title(f'A={self.A.raw_value}, F={self.f.raw_value}, P={self.p.raw_value}')
        p = axis.plot(time, self(time), **kwargs)
        return p

#%%
# Single Example
# **************
# To include embedded rST, use a line of >= 20 ``#``'s or ``#%%`` between your
# rST and your code. This separates your example
# into distinct text and code blocks. You can continue writing code below the
# embedded rST text block:

p1 = Pendulum.from_pars()
# Another pendulum with Amplitude = 5
p2 = Pendulum.from_pars(A=5)
# Another pendulum with Frequency = 4
p3 = Pendulum.from_pars(A=5, f=4)
# Another pendulum with Phase = pi/2
p4 = Pendulum.from_pars(A=5, f=4, p=np.pi/2)

#%%
# Plotting

t = np.linspace(0, 3, 601)
fig = plt.figure()
gs = fig.add_gridspec(2, 2)
(ax1, ax2), (ax3, ax4) = gs.subplots(sharex='col', sharey='row')
p1.plot(t, axis=ax1)
p2.plot(t, axis=ax2)
p3.plot(t, axis=ax3)
p4.plot(t, axis=ax4)
fig.show()

#%%
# Multiple Examples
# *****************
# To include embedded rST, use a line of >= 20 ``#``'s or ``#%%`` between your
# rST and your code. This separates your example
# into distinct text and code blocks. You can continue writing code below the
# embedded rST text block:

pendulum_array = [Pendulum.from_pars(p=phase) for phase in np.linspace(0, 1, 3)]
fig = plt.figure()
for pendulum in pendulum_array:
    pendulum.plot(t, label=f'Phase = {pendulum.p}')
plt.legend(loc='lower right')
fig.show()




# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import numpy as np
from easyscience.Objects.Groups import BaseCollection
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.Variable import Descriptor
from easyscience.Objects.Variable import Parameter

from .AtomicDisplacement import AtomicDisplacement
from .Lattice import PeriodicLattice
from .Specie import Specie
from .Susceptibility import MagneticSusceptibility

if TYPE_CHECKING:
    from easyscience.Utils.typing import iF


_SITE_DETAILS = {
    'label': {
        'value': 'H',
        'description': 'A unique identifier for a particular site in the crystal',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
    },
    'position': {
        'value': 0.0,
        'description': 'Atom-site coordinate as fractions of the unit cell length.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'fixed': True,
    },
    'occupancy': {
        'value': 1.0,
        'description': 'The fraction of the atom type present at this site.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'fixed': True,
    },
}

S = TypeVar('S', bound='Site')


class Site(BaseObj):
    label: ClassVar[Descriptor]
    specie: ClassVar[Specie]
    occupancy: ClassVar[Parameter]
    fract_x: ClassVar[Parameter]
    fract_y: ClassVar[Parameter]
    fract_z: ClassVar[Parameter]

    def __init__(
        self,
        label: Optional[Union[str, Descriptor]] = None,
        specie: Optional[Union[str, Specie]] = None,
        occupancy: Optional[Union[float, Parameter]] = None,
        fract_x: Optional[Union[float, Parameter]] = None,
        fract_y: Optional[Union[float, Parameter]] = None,
        fract_z: Optional[Union[float, Parameter]] = None,
        interface: Optional[iF] = None,
        **kwargs,
    ):
        adp = kwargs.get('adp', None)
        b_iso_or_equiv = kwargs.get('b_iso_or_equiv', None)
        u_iso_or_equiv = kwargs.get('u_iso_or_equiv', None)

        if b_iso_or_equiv is not None and u_iso_or_equiv is not None:
            raise AttributeError('Cannot set both Biso and Uiso')

        if adp is None and b_iso_or_equiv is None and u_iso_or_equiv is None:
            adp = AtomicDisplacement('Uiso', Uiso=0)
            kwargs['adp'] = adp

        if b_iso_or_equiv is not None:
            adp = AtomicDisplacement('Biso', Biso=b_iso_or_equiv)
            kwargs['adp'] = adp

        u_iso_or_equiv = kwargs.get('u_iso_or_equiv', None)

        if u_iso_or_equiv is not None:
            aadp = AtomicDisplacement('Uiso', Uiso=u_iso_or_equiv)
            kwargs['adp'] = aadp

        msp = kwargs.get('msp', None)
        if msp is not None:
            if isinstance(msp, str):
                msp = MagneticSusceptibility(msp)
            for parameter in msp.get_parameters():
                if parameter.name in kwargs.keys():
                    new_option = kwargs.pop(parameter.name)
                    parameter.value = new_option
            kwargs['msp'] = msp

        if adp is not None:
            if isinstance(adp, str):
                adp = AtomicDisplacement(adp)
            for parameter in adp.get_parameters():
                if parameter.name in kwargs.keys():
                    new_option = kwargs.pop(parameter.name)
                    parameter.value = new_option
            kwargs['adp'] = adp

        super(Site, self).__init__(
            'site',
            label=Descriptor('label', **_SITE_DETAILS['label']),
            specie=Specie(_SITE_DETAILS['label']['value']),
            occupancy=Parameter('occupancy', **_SITE_DETAILS['occupancy']),
            fract_x=Parameter('fract_x', **_SITE_DETAILS['position']),
            fract_y=Parameter('fract_y', **_SITE_DETAILS['position']),
            fract_z=Parameter('fract_z', **_SITE_DETAILS['position']),
            **kwargs,
        )
        if label is not None:
            self.label = label
        if specie is not None:
            self.specie = specie
        else:
            if label is not None:
                self.specie = label
        if occupancy is not None:
            self.occupancy = occupancy
        if fract_x is not None:
            self.fract_x = fract_x
        if fract_y is not None:
            self.fract_y = fract_y
        if fract_z is not None:
            self.fract_z = fract_z
        self.interface = interface

    def __repr__(self) -> str:
        return (
            f'Atom {self.name} ({self.specie.raw_value}) @'
            f' ({self.fract_x.raw_value}, {self.fract_y.raw_value}, {self.fract_z.raw_value})'
        )

    @property
    def name(self) -> str:
        return self.label.raw_value

    @property
    def fract_coords(self) -> np.ndarray:
        """
        Get the current sites fractional co-ordinates as an array

        :return: Array containing fractional co-ordinates
        :rtype: np.ndarray
        """
        return np.array([self.fract_x.raw_value, self.fract_y.raw_value, self.fract_z.raw_value])

    def fract_distance(self, other_site: S) -> float:
        """
        Get the distance between two sites

        :param other_site: Second site
        :param other_site: Second site
        :type other_site: Site
        :return: Distance between 2 sites
        :rtype: float
        """
        return np.linalg.norm(other_site.fract_coords - self.fract_coords)

    @property
    def x(self) -> Parameter:
        return self.fract_x

    @property
    def y(self) -> Parameter:
        return self.fract_y

    @property
    def z(self) -> Parameter:
        return self.fract_z

    @property
    def b_iso_or_equiv(self) -> Parameter:
        if not hasattr(self, 'adp'):
            return None
        if not hasattr(self.adp, 'Biso'):
            return None
        return self.adp.Biso

    @property
    def u_iso_or_equiv(self) -> Parameter:
        if not hasattr(self, 'adp'):
            return None
        if not hasattr(self.adp, 'Uiso'):
            return None
        return self.adp.Uiso

    @property
    def is_magnetic(self) -> bool:
        return getattr(self.specie, 'spin', None) is not None or hasattr(self, 'msp')

    def add_adp(self, adp_type: Union[str, AtomicDisplacement], **kwargs):
        if isinstance(adp_type, str):
            adp_type = AtomicDisplacement(adp_type, **kwargs)
        self._add_component('adp', adp_type)
        if self.interface is not None:
            self.interface.generate_bindings(self)

    def add_msp(self, msp_type: Union[str, MagneticSusceptibility], **kwargs):
        if isinstance(msp_type, str):
            msp_type = MagneticSusceptibility(msp_type, **kwargs)
        self._add_component('msp', msp_type)
        # if self.interface is not None:
        #    self.interface.generate_bindings(self)


class PeriodicSite(Site):
    def __init__(
        self,
        lattice: Optional[PeriodicLattice] = None,
        label: Optional[Union[str, Descriptor]] = None,
        specie: Optional[Union[str, Specie]] = None,
        occupancy: Optional[Union[float, Parameter]] = None,
        fract_x: Optional[Union[float, Parameter]] = None,
        fract_y: Optional[Union[float, Parameter]] = None,
        fract_z: Optional[Union[float, Parameter]] = None,
        interface: Optional[iF] = None,
        **kwargs,
    ):
        super(PeriodicSite, self).__init__(label, specie, occupancy, fract_x, fract_y, fract_z, **kwargs)
        if lattice is None:
            lattice = PeriodicLattice()
        self.lattice = lattice
        self.interface = interface

    @staticmethod
    def _from_site_kwargs(lattice: PeriodicLattice, site: S) -> Dict[str, float]:
        return {
            'lattice': lattice,
            'label': site.label,
            'specie': site.specie,
            'occupancy': site.occupancy,
            'fract_x': site.fract_x,
            'fract_y': site.fract_y,
            'fract_z': site.fract_z,
            'interface': site.interface,
        }

    @classmethod
    def from_site(cls, lattice: PeriodicLattice, site: S) -> S:
        kwargs = cls._from_site_kwargs(lattice, site)
        if hasattr(site, 'adp'):
            kwargs['adp'] = site.adp
        if hasattr(site, 'msp'):
            kwargs['msp'] = site.msp
        return cls(**kwargs)

    def get_orbit(self) -> np.ndarray:
        """
        Generate all orbits for a given fractional position.

        """
        sym_op = self.lattice.spacegroup._sg_data.get_orbit
        return sym_op(self.fract_coords)

    @property
    def cart_coords(self) -> np.ndarray:
        """
        Get the atomic position in Cartesian form.
        :return:
        :rtype:
        """
        return self.lattice.get_cartesian_coords(self.fract_coords)


class Atoms(BaseCollection):
    _SITE_CLASS = Site

    def __init__(self, name: str, *args, interface: Optional[iF] = None, **kwargs):
        if not isinstance(name, str):
            raise TypeError('A `name` for this collection must be given in string form')
        super(Atoms, self).__init__(name, *args, **kwargs)
        self.interface = interface
        self._kwargs._stack_enabled = True

    def __repr__(self) -> str:
        return f'Collection of {len(self)} sites.'

    def __getitem__(self, idx: Union[int, slice, str]) -> Union[Parameter, Descriptor, BaseObj, 'BaseCollection']:
        if isinstance(idx, str) and idx in self.atom_labels:
            idx = self.atom_labels.index(idx)
        return super(Atoms, self).__getitem__(idx)

    def __delitem__(self, key: Union[int, str]):
        if isinstance(key, str) and key in self.atom_labels:
            key = self.atom_labels.index(key)
        return super(Atoms, self).__delitem__(key)

    def remove(self, key: Union[int, str]):
        self.__delitem__(key)

    def append(self, *args, **kwargs):
        """
        Add an atom to the crystal
        """
        if len(args) == 1 and isinstance(args[0], Site):
            atom = args[0]
        else:
            atom = Site(*args, **kwargs)
        super(Atoms, self).append(atom)

    # def append(self, item: S):
    #     if not issubclass(type(item), Site):
    #         raise TypeError("Item must be a Site")
    #     if item.label.raw_value in self.atom_labels:
    #         raise AttributeError(
    #             f"An atom of name {item.label.raw_value} already exists."
    #         )
    #     super(Atoms, self).append(item)

    @property
    def atom_labels(self) -> List[str]:
        return [atom.label.raw_value for atom in self]

    @property
    def atom_species(self) -> List[str]:
        return [atom.specie.raw_value for atom in self]

    @property
    def atom_occupancies(self) -> np.ndarray:
        return np.array([atom.occupancy.raw_value for atom in self])


A = TypeVar('A', bound=Atoms)


class PeriodicAtoms(Atoms):
    _SITE_CLASS = PeriodicSite

    def __init__(
        self,
        name: str,
        *args,
        lattice: Optional[PeriodicLattice] = None,
        interface: Optional[iF] = None,
        **kwargs,
    ):
        args = list(args)
        if lattice is None:
            for item in args:
                if hasattr(item, 'lattice'):
                    lattice = item.lattice
                    break
        if lattice is None:
            raise AttributeError
        for idx, item in enumerate(args):
            if issubclass(type(item), Site):
                args[idx] = self._SITE_CLASS.from_site(lattice, item)
        super(PeriodicAtoms, self).__init__(name, *args, **kwargs, interface=interface)
        self.lattice = lattice

    @classmethod
    def from_atoms(cls, lattice: PeriodicLattice, atoms: Atoms) -> A:
        return cls(atoms.name, *atoms, lattice=lattice, interface=atoms.interface)

    def __repr__(self) -> str:
        return f'Collection of {len(self)} periodic sites.'

    def append(self, item: S):
        if not issubclass(item.__class__, Site):
            raise TypeError('Item must be a Site or periodic site')
        if item.label.raw_value in self.atom_labels:
            raise AttributeError(f'An atom of name {item.label.raw_value} already exists.')
        # if isinstance(item, Site):
        item = self._SITE_CLASS.from_site(self.lattice, item)
        super(PeriodicAtoms, self).append(item)

    def get_orbits(self, magnetic_only: bool = False):
        orbit_dict = {}
        for item in self:
            if magnetic_only and not item.is_magnetic:
                continue
            orbit_dict[item.label.raw_value] = item.get_orbit()
        return orbit_dict



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

import pytest
import easyscience
import numpy as np
from numbers import Number
from easycrystallography.Components.Lattice import Lattice, Parameter, CELL_DETAILS
from easyscience import global_object

pars_dict = {
    "cubic": (5, 5, 5, 90, 90, 90),
    "tetragonal": (10, 10, 5, 90, 90, 90),
    "orthorhombic": (2, 3, 4, 90, 90, 90),
    "monoclinic": (2, 3, 4, 90, 99, 90),
    "hexagonal": (3, 3, 4, 90, 90, 120),
    "rhombohedral": (4, 4, 4, 99, 99, 99),
}


def mod_pars(in_mods=None, sep=False) -> tuple:
    items = []
    keys = pars_dict.keys()
    if in_mods is None:
        in_mods = [[]] * len(keys)
    for key, mod_ in zip(keys, in_mods):
        if mod_:
            if sep:
                items.append(pytest.param(pars_dict[key], mod_, id=key))
            else:
                items.append(pytest.param((*pars_dict[key], mod_), id=key))
        else:
            items.append(pytest.param(pars_dict[key], id=key))
    return tuple(items)


basic_pars = mod_pars()

matrix_pars = mod_pars(
    [
        [
            [5.000000e00, 0.000000e00, 3.061617e-16],
            [-3.061617e-16, 5.000000e00, 3.061617e-16],
            [0.000000e00, 0.000000e00, 5.000000e00],
        ],
        [
            [1.000000e01, 0.000000e00, 6.123234e-16],
            [-6.123234e-16, 1.000000e01, 6.123234e-16],
            [0.000000e00, 0.000000e00, 5.000000e00],
        ],
        [
            [2.0000000e00, 0.0000000e00, 1.2246468e-16],
            [-1.8369702e-16, 3.0000000e00, 1.8369702e-16],
            [0.0000000e00, 0.0000000e00, 4.0000000e00],
        ],
        [
            [1.97537668e00, 0.00000000e00, -3.12868930e-01],
            [-1.83697020e-16, 3.00000000e00, 1.83697020e-16],
            [0.00000000e00, 0.00000000e00, 4.00000000e00],
        ],
        [
            [3.00000000e00, 0.00000000e00, 1.83697020e-16],
            [-1.50000000e00, 2.59807621e00, 1.83697020e-16],
            [0.00000000e00, 0.00000000e00, 4.00000000e00],
        ],
        [
            [3.95075336, 0.0, -0.62573786],
            [-0.7326449, 3.88222663, -0.62573786],
            [0.0, 0.0, 4.0],
        ],
    ]
)


def test_Lattice_default():
    lattice = Lattice()

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for key in items:
        item: Parameter = getattr(lattice, key)
        assert item.name == key
        t = key.split("_")[0]
        test_defaults = CELL_DETAILS[t].copy()
        test_defaults["raw_value"] = test_defaults["value"]
        del test_defaults["value"]
        test_defaults["unit"] = test_defaults["units"]
        del test_defaults["units"]
        for default in test_defaults.keys():
            r = test_defaults[default]
            i = getattr(item, default)
            if default == "unit":
                i = str(i)[0:3]
                r = r[0:3]
            assert i == r


@pytest.mark.parametrize("ang_unit", ("deg", "rad"))
@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_from_pars(value: list, ang_unit: str):
    ref = [v for v in value]

    if ang_unit == "rad":
        value = [
            value[0],
            value[1],
            value[2],
            np.deg2rad(value[3]),
            np.deg2rad(value[4]),
            np.deg2rad(value[5]),
        ]

    l = Lattice(*value, ang_unit=ang_unit)

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for idx, key in enumerate(items):
        item: Parameter = getattr(l, key)
        assert item.name == key
        t = key.split("_")[0]
        test_defaults = CELL_DETAILS[t].copy()
        test_defaults["raw_value"] = ref[idx]
        del test_defaults["value"]
        test_defaults["unit"] = test_defaults["units"]
        del test_defaults["units"]
        for default in test_defaults.keys():
            r = test_defaults[default]
            i = getattr(item, default)
            if default == "unit":
                i = str(i)[0:3]
                r = r[0:3]
            if isinstance(i, Number) and not isinstance(i, bool):
                assert r == pytest.approx(i)
            else:
                assert i == r

#FAILED tests/unit_tests/Components/test_Lattice.py::test_Lattice_from_matrix[monoclinic] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
@pytest.mark.parametrize("value", matrix_pars)
def test_Lattice_from_matrix(value):
    global_object.map._clear()
    args = value[0:-1]
    matrix = value[-1]
    l = Lattice.from_matrix(matrix)
    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]
    for idx, key in enumerate(items):
        item: Parameter = getattr(l, key)
        assert item.name == key
        t = key.split("_")[0]
        test_defaults = CELL_DETAILS[t].copy()
        test_defaults["raw_value"] = args[idx]
        del test_defaults["value"]
        test_defaults["unit"] = test_defaults["units"]
        del test_defaults["units"]
        for default in test_defaults.keys():
            r = test_defaults[default]
            i = getattr(item, default)
            if default == "unit":
                i = str(i)[0:3]
                r = r[0:3]
            if isinstance(i, Number) and not isinstance(i, bool):
                assert r == pytest.approx(i)
            else:
                assert i == r


@pytest.mark.parametrize(
    "value", mod_pars([[0], [0, 2], [0, 1, 2], [0, 1, 2, 4], [0, 2], [0, 3]])
)
def test_Lattice_from_special(request, value):
    ref = np.array(value[0:6])
    cons = ref[value[6:]]
    lattice_type = request.node.name.split("[")[1][:-1]

    f = getattr(Lattice, lattice_type)
    l = f(*cons)

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for idx, key in enumerate(items):
        item: Parameter = getattr(l, key)
        assert item.name == key
        t = key.split("_")[0]
        test_defaults = CELL_DETAILS[t].copy()
        test_defaults["raw_value"] = ref[idx]
        del test_defaults["value"]
        test_defaults["unit"] = test_defaults["units"]
        del test_defaults["units"]
        for default in test_defaults.keys():
            r = test_defaults[default]
            i = getattr(item, default)
            if default == "unit":
                i = str(i)[0:3]
                r = r[0:3]
            if isinstance(i, Number) and not isinstance(i, bool):
                assert r == pytest.approx(i)
            else:
                assert i == r


@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_pars_short_GET(value: list):
    l = Lattice(*value)

    items = ["a", "b", "c", "alpha", "beta", "gamma"]

    for idx, item in enumerate(items):
        f = getattr(l, item)
        assert f == value[idx]


@pytest.mark.parametrize(
    "in_value, new_value",
    mod_pars(
        [
            (6, 6, 6, 90, 90, 90),
            (11, 11, 6, 90, 90, 90),
            (5, 6, 7, 90, 90, 90),
            (6, 7, 8, 90, 95, 90),
            (4, 4, 5, 90, 90, 120),
            (6, 6, 6, 95, 95, 95),
        ],
        True,
    ),
)
def test_Lattice_pars_short_SET(in_value: list, new_value: list):
    l = Lattice(*in_value)

    items = ["a", "b", "c", "alpha", "beta", "gamma"]

    for idx, item in enumerate(items):
        f = getattr(l, item)
        assert f == in_value[idx]
        setattr(l, item, new_value[idx])
        f = getattr(l, item)
        assert f == new_value[idx]


@pytest.mark.parametrize(
    "in_value, new_value",
    mod_pars(
        [
            (6, 6, 6, 90, 90, 90),
            (11, 11, 6, 90, 90, 90),
            (5, 6, 7, 90, 90, 90),
            (6, 7, 8, 90, 95, 90),
            (4, 4, 5, 90, 90, 120),
            (6, 6, 6, 95, 95, 95),
        ],
        True,
    ),
)
def test_Lattice_pars_SET(in_value: list, new_value: list):
    l = Lattice(*in_value)

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for idx, item in enumerate(items):
        f = getattr(l, item)
        assert f.raw_value == in_value[idx]
        setattr(l, item, new_value[idx])
        f = getattr(l, item)
        assert f.raw_value == new_value[idx]


@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_angles(value: list):
    l = Lattice(*value)
    assert np.all(np.array(value[3:]) == l.angles)


@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_lengths(value: list):
    l = Lattice(*value)
    assert np.all(np.array(value[0:3]) == l.lengths)


@pytest.mark.parametrize("value", matrix_pars)
def test_Lattice_matrix(value: list):
    args = value[0:-1]
    matrix = np.array(value[-1])

    l = Lattice(*args)
    assert np.all(np.isclose(matrix, l.matrix))


@pytest.mark.parametrize("value", matrix_pars)
def test_Lattice_inv_matrix(value: list):
    args = value[0:-1]
    matrix = np.array(value[-1])
    matrix = np.linalg.inv(matrix)

    l = Lattice(*args)
    assert np.all(np.isclose(matrix, l.inv_matrix))


@pytest.mark.parametrize(
    "value",
    mod_pars(
        [125.0, 500.0, 24.0, 23.704520174283306, 31.1769145362398, 61.35087958926781]
    ),
)
def test_Lattice_volume(value):
    args = value[:-1]
    volume = value[-1]

    l = Lattice(*args)
    assert volume == pytest.approx(l.volume.m)
    assert str(l.volume.units) == "angstrom ** 3"


@pytest.mark.parametrize("value", matrix_pars)
def test_Lattice_metric_tensor(value):
    args = value[0:-1]

    matrix = np.array(value[-1])
    matrix = np.dot(matrix, matrix.T)

    l = Lattice(*args)
    assert np.all(np.isclose(matrix, l.metric_tensor))


@pytest.mark.parametrize(
    "in_value, new_value",
    mod_pars(
        [
            (1.256637, 1.256637, 1.256637, 90, 90, 90),
            (0.6283185, 0.6283185, 1.256637, 90, 90, 90),
            (3.14159, 2.09439510, 1.5707963, 90, 90, 90),
            (3.180753, 2.0943951, 1.5903765, 90, 81, 90),
            (2.418399, 2.418399, 1.570796, 90, 90, 60),
            (1.61845, 1.61845, 1.61845, 79.31296, 79.31296, 79.31296),
        ],
        True,
    ),
)
def test_Lattice_reciprocal_lattice(in_value: list, new_value: list):
    l = Lattice(*in_value)
    obj = l.reciprocal_lattice

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for idx, item in enumerate(items):
        f = getattr(obj, item)
        assert np.isclose(f.raw_value, new_value[idx])


@pytest.mark.parametrize(
    "in_value, new_value",
    mod_pars(
        [
            (0.2, 0.2, 0.2, 90, 90, 90),
            (0.1, 0.1, 0.2, 90, 90, 90),
            (0.5, 1 / 3, 0.25, 90, 90, 90),
            (0.5062325, 1 / 3, 0.253116, 90, 81, 90),
            (0.3849, 0.3849, 0.25, 90, 90, 60),
            (0.257584, 0.257584, 0.257584, 79.31296, 79.31296, 79.31296),
        ],
        True,
    ),
)
def test_Lattice_reciprocal_lattice(in_value: list, new_value: list):
    l = Lattice(*in_value)
    obj = l.reciprocal_lattice_crystallographic

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for idx, item in enumerate(items):
        f = getattr(obj, item)
        assert np.isclose(f.raw_value, new_value[idx])


@pytest.mark.parametrize("scale", [0.1, 2, 3.14, 100])
@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_scale(value: list, scale: float):
    l = Lattice(*value)

    new_volume = scale * l.volume
    scaled = l.scale(new_volume)

    assert np.isclose(scaled.volume, new_volume)
    assert np.all(np.isclose(l.angles, scaled.angles))


@pytest.mark.parametrize(
    "scale", [0.1, 2, 3.14, 100, [0.5, 0.5, 1], [0.5, 1, 0.5], [1, 0.5, 0.5]]
)
@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_length_scale(value: list, scale: float):
    l = Lattice(*value)

    scaled = l.scale_lengths(scale)
    assert np.all(np.isclose(l.angles, scaled.angles))
    assert np.all(np.isclose(np.array(l.lengths) * scale, scaled.lengths))


@pytest.mark.parametrize(
    "co_ords", [[0.1, 2, 3.14], [0.5, 0.5, 1], [0.5, 1, 0.5], [1, 0.5, 0.5]]
)
@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_fract_cart_coords(value: list, co_ords: list):
    l = Lattice(*value)

    frac = l.get_fractional_coords(co_ords)
    cart_co_ords = l.get_cartesian_coords(frac)

    assert np.all(np.isclose(co_ords, cart_co_ords))


@pytest.mark.parametrize("crystal_system", ["cubic", "tetragonal", "orthorhombic"])
def test_Lattice_is_orthogonal(crystal_system):

    l = Lattice(*pars_dict[crystal_system])
    assert l.is_orthogonal()


def test_Lattice_is_hexagonal():

    lengths = np.array(pars_dict["hexagonal"][0:3])
    angles = np.array(pars_dict["hexagonal"][3:])

    l = Lattice(*lengths, *angles)
    assert l.is_hexagonal()

    l = Lattice(*lengths, *(angles + 1e-5))
    assert l.is_hexagonal(hex_angle_tol=1e-4)
    assert not l.is_hexagonal(hex_angle_tol=1e-6)

    l = Lattice(*(lengths + [1e-2, -1e-2, 1e-2]), *angles)
    assert l.is_hexagonal(hex_length_tol=1e-1)
    assert not l.is_hexagonal(hex_length_tol=1e-7)


@pytest.mark.parametrize(
    "values, out_str",
    mod_pars(
        [
            "<Lattice: (a: 5.00 Å, b: 5.00 Å, c: 5.00Å, alpha: 90.00 deg, beta: 90.00 deg, gamma: 90.00 deg>",
            "<Lattice: (a: 10.00 Å, b: 10.00 Å, c: 5.00Å, alpha: 90.00 deg, beta: 90.00 deg, gamma: 90.00 deg>",
            "<Lattice: (a: 2.00 Å, b: 3.00 Å, c: 4.00Å, alpha: 90.00 deg, beta: 90.00 deg, gamma: 90.00 deg>",
            "<Lattice: (a: 2.00 Å, b: 3.00 Å, c: 4.00Å, alpha: 90.00 deg, beta: 99.00 deg, gamma: 90.00 deg>",
            "<Lattice: (a: 3.00 Å, b: 3.00 Å, c: 4.00Å, alpha: 90.00 deg, beta: 90.00 deg, gamma: 120.00 deg>",
            "<Lattice: (a: 4.00 Å, b: 4.00 Å, c: 4.00Å, alpha: 99.00 deg, beta: 99.00 deg, gamma: 99.00 deg>",
        ],
        True,
    ),
)
def test_Lattice_repr(values, out_str):
    l = Lattice(*values)
    assert str(l) == out_str


def make_dict(value) -> dict:
    return {
        "@module": "easycrystallography.Components.Lattice",
        "@class": "Lattice",
        "length_a": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "length_a",
            "value": float(value[0]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell length of the selected structure in angstroms.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "angstrom",
            "unique_name": "149027786693506016496254445195239597714",
            "enabled": True,
        },
        "length_b": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "length_b",
            "value": float(value[1]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell length of the selected structure in angstroms.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "angstrom",
            "unique_name": "294836968667493729920294930317191977696",
            "enabled": True,
        },
        "length_c": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "length_c",
            "value": float(value[2]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell length of the selected structure in angstroms.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "angstrom",
            "unique_name": "275642519607899521714432039990092728990",
            "enabled": True,
        },
        "angle_alpha": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "angle_alpha",
            "value": float(value[3]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell angle of the selected structure in degrees.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "degree",
            "unique_name": "161899496656810433045540450883723049023",
            "enabled": True,
        },
        "angle_beta": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "angle_beta",
            "value": float(value[4]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell angle of the selected structure in degrees.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "degree",
            "unique_name": "186637124621565458307862080073460500737",
            "enabled": True,
        },
        "angle_gamma": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "name": "angle_gamma",
            "value": float(value[5]),
            "error": 0.0,
            "min": 0,
            "max": np.inf,
            "fixed": True,
            "description": "Unit-cell angle of the selected structure in degrees.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_cell/",
            "units": "degree",
            "unique_name": "225244117838730303286513043607480352526",
            "enabled": True,
        },
        "interface": None,
        "unique_name": "78109834334085432621980127205750673524",
    }


@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_as_dict(value: list):
    l = Lattice(*value)
    obtained = l.as_dict()
    expected = make_dict(value)

    def check_dict(check, item):
        if isinstance(check, dict) and isinstance(item, dict):
            for this_check_key in check.keys():
                if this_check_key == "unique_name":
                    continue
                check_dict(check[this_check_key], item[this_check_key])
        else:
            assert isinstance(item, type(check))
            assert item == check

    check_dict(expected, obtained)


@pytest.mark.parametrize("value", basic_pars)
def test_Lattice_from_dict(value: list):
    # ADDED UNTIL UNIQUE_NAME IS FIXED BECAUSE OF THE FOLLOWING ERRORS:
    # FAILED tests/unit_tests/Components/test_Lattice.py::test_Lattice_from_dict[tetragonal] - ValueError: Object name 149027786693506016496254445195239597714 already exists in the graph.
    # FAILED tests/unit_tests/Components/test_Lattice.py::test_Lattice_from_dict[monoclinic] - ValueError: Object name 149027786693506016496254445195239597714 already exists in the graph.
    # FAILED tests/unit_tests/Components/test_Lattice.py::test_Lattice_from_dict[rhombohedral] - ValueError: Object name 149027786693506016496254445195239597714 already exists in the grap
    global_object.map._clear()

    expected = make_dict(value)
    l = Lattice.from_dict(expected)
    obtained = l.as_dict()

    def check_dict(check, item):
        if isinstance(check, dict) and isinstance(item, dict):
            for this_check_key in check.keys():
                if this_check_key == "unique_name":
                    continue
                check_dict(check[this_check_key], item[this_check_key])
        else:
            assert isinstance(item, type(check))
            assert item == check

    check_dict(expected, obtained)


@pytest.mark.parametrize("value", basic_pars)
@pytest.mark.parametrize("fmt", [".3f"])
@pytest.mark.parametrize("opt", ["m", "l", "t"])
def test_Lattice_fmt(value, fmt, opt):

    l = Lattice(*value)
    out_fmt = "{}{}".format(fmt, opt)

    def do_test(in_str):
        m = (l.lengths, l.angles)
        if opt == "m":
            m = l.matrix.tolist()
            fmt2 = "[[{}, {}, {}], [{}, {}, {}], [{}, {}, {}]]"
        elif opt == "l":
            fmt2 = "{{{}, {}, {}, {}, {}, {}}}"
        else:
            fmt2 = "({} {} {}), ({} {} {})"
        check_str = fmt2.format(*[format(c, fmt) for row in m for c in row])
        assert in_str == check_str

    # Ancient Python. We won't be supporting this
    with pytest.raises(TypeError):
        out_fmt2 = "%" + out_fmt
        out_str = out_fmt2 % l
        do_test(out_str)
    # Python >2.7 "{:03fm}".format(l)
    out_fmt2 = "{:" + f"{out_fmt}" + "}"
    out_str = out_fmt2.format(l)
    do_test(out_str)
    # Python >3.6 + f"{l:03fm}"
    # This is stupidly dangerous.
    # Releases dragons, orks and is where darkness lies. You've been warned
    # !!!! DO NOT USE OUT OF THIS UNIQUE CONTEXT !!!!

    def effify(non_f_str: str, l: Lattice) -> str:
        return eval(f'f"""{non_f_str}"""')

    out_str = effify(f"{{l:{out_fmt}}}", l)
    do_test(out_str)


@pytest.mark.parametrize("value", basic_pars)
def test_lattice_copy(value):
    from copy import copy

    l1 = Lattice(*value)
    l2 = copy(l1)

    items = [
        "length_a",
        "length_b",
        "length_c",
        "angle_alpha",
        "angle_beta",
        "angle_gamma",
    ]

    for item in items:
        f1 = getattr(l1, item)
        f2 = getattr(l2, item)
        assert np.isclose(f1.raw_value, f2.raw_value)
        assert f1 != f2



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

import textwrap
from collections.abc import MutableMapping
from math import floor
from math import log10
from numbers import Number
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
from easyscience.Utils.io.dict import DataDictSerializer
from easyscience.Utils.io.dict import DictSerializer
from easyscience.Utils.io.template import BaseEncoderDecoder
from gemmi import cif

if TYPE_CHECKING:
    from easyscience.Utils.typing import BV

_MAX_LEN = 160
_SEP = '.'


def _flatten_dict_gen(d, parent_key, sep):
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            yield from flatten_dict(v, new_key, sep=sep).items()
        elif isinstance(v, list):
            yield (
                new_key,
                [value if not isinstance(value, MutableMapping) else flatten_dict(value, '', sep) for value in v],
            )
        else:
            yield new_key, v


def flatten_dict(d: MutableMapping, parent_key: str = '', sep: str = '.') -> dict:
    return dict(_flatten_dict_gen(d, parent_key, sep))


def _unflatten_dict(key, value, out, sep) -> dict:
    key, *rest = key.split(sep, 1)
    if rest:
        _unflatten_dict(rest[0], value, out.setdefault(key, {}), sep)
    else:
        v, e, f = CifSerializer.string_to_variable(value)
        out[key] = v
        if e is not None:
            out['error'] = e
        if f is not None and not f:
            out['fixed'] = f


def _unflatten_loop(loop: cif.Loop, out: dict, sep: str):
    loop_name = ''.join(p for p, *r in zip(*loop.tags) if all(p == c for c in r))
    labels = [lab[len(loop_name) :] for lab in loop.tags]
    out[loop_name] = []
    width = loop.width()
    length = loop.length()
    for idx in range(length):
        this_dict = {}
        for idx2 in range(width):
            _unflatten_dict(labels[idx2], loop.val(idx, idx2), this_dict, sep)
        out[loop_name].append(this_dict)
    return out


def unflatten_bock(block: cif.Block, sep: str) -> dict:
    # Note that all names start with "_", so start at index 1!
    inter_dict = {item.pair[0][1:]: item.pair[1] for item in block if item.loop is None}
    loops = [loop.loop for loop in block if loop.loop is not None]
    out_dict = {}
    for k, v in inter_dict.items():
        _unflatten_dict(k, v, out_dict, sep)
    for loop in loops:
        _unflatten_loop(loop, out_dict, sep)
    return out_dict


class CifSerializer(BaseEncoderDecoder):
    def encode(self, obj: BV, skip: Optional[List[str]] = None, data_only: bool = False) -> str:
        if skip is None:
            skip = []
        doc = cif.Document()
        encoder = DictSerializer
        if data_only:
            encoder = DataDictSerializer
        obj_dict = obj.encode(encoder=encoder, skip=skip, include_id=False)
        flattened_obj = flatten_dict(obj_dict, sep=_SEP)
        block = doc.add_new_block(str(obj._borg.map.convert_id_to_key(obj)))
        for k, v in flattened_obj.items():
            self._check_class(doc, block, k, v)
        return doc.as_string()

    @classmethod
    def decode(cls, data_str) -> Union[BV, Dict[str, BV]]:
        document = cif.read_string(data_str)
        out_dict = {}
        for i, block in enumerate(document):
            out_dict[block.name] = DictSerializer.decode(unflatten_bock(block, sep=_SEP))
        if len(document) == 1:
            out_dict = list(out_dict.values())[0]
        return out_dict

    def _check_class(self, document: cif.Document, block: cif.Block, key: str, value: Any):
        T_ = type(value)
        if issubclass(T_, bool):
            new_obj = '@bool.' + str(value)
            self._str_encoder(block, key, new_obj)
        elif issubclass(T_, list):
            if len(value) > 0:
                test_item = value[0]
                if not issubclass(type(test_item), dict):
                    raise NotImplementedError
                else:
                    names = list(test_item.keys())
                    self._list_encoder(block, key, value, names)
        elif issubclass(T_, (str, Number)):
            self._str_encoder(block, key, value)

    def _str_encoder(self, block: cif.Block, key: str, in_str: str):
        in_str = self.variable_to_string(in_str)
        if key[0] != '_':
            key = '_' + key
        block.set_pair(key, in_str)

    def _list_encoder(self, block: cif.Block, name: str, items: List, keys: List[str]):
        if name[0] != '_':
            name = '_' + name
        loop: cif.Loop = block.init_loop(name, keys)
        for item in items:
            if len(keys) == 1:
                loop.add_row([self.variable_to_string(item)])
            else:
                loop.add_row([self.variable_to_string(item.get(key)) for key in keys])

    def variable_to_string(self, value, fixed: Optional[bool] = None, error: Optional[float] = None) -> str:
        decimal_places = 8
        if isinstance(value, Number):
            try:
                if np.isinf(value):
                    return self.variable_to_string(str(value))
            except TypeError:
                print(type(value))
            if abs(value - int(value)) > 0.0:
                decimal_places = len(str(value).split('.')[1]) - 1
            initial_str = '{:.' + str(decimal_places) + 'f}'
            s = initial_str.format(round(value, decimal_places))
            if error is not None and not np.isclose(error, 0.0):
                xe_exp = int(floor(log10(error)))

                # uncertainty
                un_exp = xe_exp - decimal_places + 1
                un_int = round(error * 10 ** (-un_exp))

                # nominal value
                no_exp = un_exp
                no_int = round(value * 10 ** (-no_exp))

                # format - nom(unc)
                fmt = '%%.%df' % max(0, -no_exp)
                s = (fmt + '(%.0f)') % (
                    no_int * 10**no_exp,
                    un_int * 10 ** max(0, un_exp),
                )
        elif isinstance(value, str):
            s = '{:s}'.format(value)
        else:
            s = '{:s}'.format(str(value))
        if fixed is not None and not fixed and error is None:
            s += '()'
        return self._format_field(s)

    @staticmethod
    def string_to_variable(in_string: str):
        in_string = in_string.strip()
        if "'" in in_string:
            in_string = in_string.replace("'", '')
        if '"' in in_string:
            in_string = in_string.replace('"', '')
        fixed = None
        error = None
        tokens = in_string.split('(')
        try:
            value = float(tokens[0])
        except ValueError:
            value = tokens[0]
            if value[0] == '@':
                tokens = value[1:].split('.')
                cls = tokens[0]
                val = tokens[1]
                value = eval(cls)(val)  # noqa S307
            elif value == 'None':
                value = None
            return value, error, fixed
        if len(tokens) > 1:
            fixed = False
            if tokens[1][0] != ')':
                error = (10 ** -(len(f'{tokens[0]}'.split('.')[1]) - 1)) * int(tokens[1][:-1])
        return value, error, fixed

    @staticmethod
    def _format_field(v):
        v = v.__str__().strip()
        if len(v) > _MAX_LEN:
            return ';\n' + textwrap.fill(v, _MAX_LEN) + '\n;'
        # add quotes if necessary
        if v == '':
            return '""'
        if (' ' in v or v[0] == '_') and not (v[0] == "'" and v[-1] == "'") and not (v[0] == '"' and v[-1] == '"'):
            if "'" in v:
                q = '"'
            else:
                q = "'"
            v = q + v + q
        return v



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
from easyscience.Objects.Groups import BaseCollection
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Descriptor
from easyscience.Objects.ObjectClasses import Parameter

from easycrystallography.Components.Lattice import Lattice
from easycrystallography.Components.Lattice import PeriodicLattice
from easycrystallography.Components.Site import Atoms
from easycrystallography.Components.Site import PeriodicAtoms
from easycrystallography.Components.Site import Site
from easycrystallography.Components.SpaceGroup import SpaceGroup
from easycrystallography.io.parser import Parsers

if TYPE_CHECKING:
    from easyscience.Utils.typing import iF


class Phase(BaseObj):
    _SITE_CLASS = Site
    _ATOMS_CLASS = Atoms

    cell = ClassVar[PeriodicLattice]
    _spacegroup = ClassVar[SpaceGroup]
    atoms = ClassVar[PeriodicAtoms]
    scale = ClassVar[Parameter]

    _REDIRECT = {
        'space_group': lambda obj: getattr(obj, '_spacegroup'),
    }

    def __init__(
        self,
        name: str,
        space_group: Optional[Union[SpaceGroup, str]] = None,
        cell: Optional[Union[Lattice, PeriodicLattice]] = None,
        atoms: Optional[Atoms] = None,
        scale: Optional[Parameter] = None,
        interface: Optional[iF] = None,
        enforce_sym: bool = True,
    ):
        self.name = name
        if space_group is None:
            space_group = SpaceGroup()
        if cell is None:
            cell = Lattice()
        if isinstance(cell, Lattice):
            cell = PeriodicLattice.from_lattice_and_spacegroup(cell, space_group)
        if atoms is None:
            atoms = Atoms('atoms')
        if scale is None:
            scale = Parameter('scale', 1, min=0)

        super(Phase, self).__init__(name, cell=cell, _spacegroup=space_group, atoms=atoms, scale=scale)
        if not enforce_sym:
            self.cell.clear_sym()
        self._enforce_sym = enforce_sym
        self.interface = interface

        self._extent = np.array([1, 1, 1])
        self._centre = np.array([0, 0, 0])
        self.atom_tolerance = 1e-4

    def add_atom(self, *args, **kwargs):
        """
        Add an atom to the crystal
        """
        supplied_atom = False
        for arg in args:
            if issubclass(arg.__class__, self._SITE_CLASS):
                self.atoms.append(arg)
                supplied_atom = True
        if not supplied_atom:
            atom = Site(*args, **kwargs)
            self.atoms.append(atom)

    def remove_atom(self, key):
        del self.atoms[key]

    def all_orbits(self, extent=None, magnetic_only: bool = False) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.

        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """

        if extent is None:
            extent = self._extent

        offsets = np.array(
            np.meshgrid(
                range(0, extent[0] + 1),
                range(0, extent[1] + 1),
                range(0, extent[2] + 1),
            )
        ).T.reshape(-1, 3)

        orbits = self.get_orbits(magnetic_only=magnetic_only)
        for orbit_key in orbits.keys():
            orbit = orbits[orbit_key]
            site_positions = np.apply_along_axis(np.add, 1, offsets, orbit).reshape((-1, 3)) - self.center
            orbits[orbit_key] = (
                site_positions[
                    np.all(site_positions >= -self.atom_tolerance, axis=1)
                    & np.all(site_positions <= extent + self.atom_tolerance, axis=1),
                    :,
                ]
                + self.center
            )
        return orbits

    def get_orbits(self, magnetic_only: bool = False) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.

        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """
        atoms = PeriodicAtoms.from_atoms(self.cell, self.atoms)
        orbits = atoms.get_orbits(magnetic_only=magnetic_only)
        return orbits

    @property
    def enforce_sym(self):
        return self._enforce_sym

    @enforce_sym.setter
    def enforce_sym(self, value: bool):
        if value:
            self.cell.enforce_sym()
        else:
            self.cell.clear_sym()

    @property
    def space_group(self):
        return self._spacegroup

    def set_spacegroup(self, value):
        if self._enforce_sym:
            self.cell.space_group_HM_name = value
        else:
            self._spacegroup.space_group_HM_name = value

    @property
    def extent(self) -> np.ndarray:
        """
        Get the current extent in unit cells

        :return: current extent in unit cells
        :rtype: np.ndarray
        """
        return self._extent

    @extent.setter
    def extent(self, new_extent: Union[list, np.ndarray]):
        """
        The current extent of in unit cells. Default (1, 1, 1)

        :param new_extent: The new extent in unit cells.
        :type new_extent: Union[list, tuple, np.ndarray]
        """
        if isinstance(new_extent, list):
            new_extent = np.array(new_extent)
        if np.prod(new_extent.shape) != 3:
            raise ValueError
        new_extent = new_extent.reshape((3,))
        self._extent = new_extent

    @property
    def center(self) -> np.ndarray:
        """
        Get the center position

        :return: center position
        :rtype: np.ndarray
        """
        return self._centre

    @center.setter
    def center(self, new_center: Union[list, tuple, np.ndarray]):
        """
        Set the center position. Default (0, 0, 0)

        :param new_center: New center position.
        :type new_center: Union[list, tuple, np.ndarray]
        """

        if isinstance(new_center, list):
            new_center = np.array(new_center)
        if np.prod(new_center.shape) != 3:
            raise ValueError
        new_center = new_center.reshape((3,))
        self._centre = new_center

    def _generate_positions(self, site, extent) -> np.ndarray:
        """
        Generate all orbits for a given fractional position.
        """
        sym_op = self.space_group._sg_data.get_orbit
        offsets = np.array(
            np.meshgrid(
                range(0, extent[0] + 1),
                range(0, extent[1] + 1),
                range(0, extent[2] + 1),
            )
        ).T.reshape(-1, 3)
        return np.apply_along_axis(np.add, 1, offsets, np.array(sym_op(site.fract_coords))).reshape((-1, 3))

    def all_sites(self, extent=None) -> Dict[str, np.ndarray]:
        """
        Generate all atomic positions from the atom array and symmetry operations over an extent.
        :return:  dictionary with keys of atom labels, containing numpy arrays of unique points in the extent
        (0, 0, 0) -> obj.extent
        :rtype: Dict[str, np.ndarray]
        """
        if self.space_group is None:
            return {atom.label: atom.fract_coords for atom in self.atoms}

        if extent is None:
            extent = self._extent

        sites = {}
        for site in self.atoms:
            unique_sites = self._generate_positions(site, extent)
            site_positions = unique_sites - self.center
            sites[site.label.raw_value] = (
                site_positions[
                    np.all(site_positions >= -self.atom_tolerance, axis=1)
                    & np.all(site_positions <= extent + self.atom_tolerance, axis=1),
                    :,
                ]
                + self.center
            )
        return sites

    @property
    def cif(self) -> str:
        s = ''
        cif_str_parser = Parsers('cif_str')
        with cif_str_parser.writer() as r:
            s += r.structure(self)
        s = s.strip()  # remove leading and trailing newlines
        return s

    @classmethod
    def from_cif_file(cls, filename):
        s = None
        with Parsers('cif').reader(filename) as r:
            s = r.structures(phase_class=cls)
        return s

    @classmethod
    def from_cif_string(cls, cif_string: str):
        s = None
        with Parsers('cif_str').reader() as r:
            s = r.structures(cif_string, phase_class=cls)
        return s


class Phases(BaseCollection):
    _SITE_CLASS = Site
    _ATOM_CLASS = Atoms
    _PHASE_CLASS = Phase

    def __init__(self, name: str = 'phases', *args, interface: Optional[iF] = None, **kwargs):
        """
        Generate a collection of crystals.

        :param name: Name of the crystals collection
        :type name: str
        :param args: objects to create the crystal
        :type args: *Phase
        """
        if not isinstance(name, str):
            raise AttributeError('Name should be a string!')

        super(Phases, self).__init__(name, *args, **kwargs)
        self.interface = interface

    def __repr__(self) -> str:
        return f'Collection of {len(self)} phases: {self.phase_names}'

    def __getitem__(self, idx: Union[int, slice]) -> Union[Parameter, Descriptor, BaseObj, BaseCollection]:
        if isinstance(idx, str) and idx in self.phase_names:
            idx = self.phase_names.index(idx)
        return super(Phases, self).__getitem__(idx)

    def __delitem__(self, key):
        if isinstance(key, str) and key in self.phase_names:
            key = self.phase_names.index(key)
        return super(Phases, self).__delitem__(key)

    def append(self, item: Phase):
        if not isinstance(item, Phase):
            raise TypeError('Item must be a Phase')
        if item.name in self.phase_names:
            raise AttributeError(f'A phase of name {item.name} already exists.')
        super(Phases, self).append(item)

    @property
    def phase_names(self) -> List[str]:
        return [phase.name for phase in self]

    @property
    def cif(self) -> str:
        s = ''
        cif_str_parser = Parsers('cif_str')
        with cif_str_parser.writer() as r:
            s += r.structures(self)
        return s

    @classmethod
    def from_cif_file(cls, filename):
        s = None
        with Parsers('cif').reader(filename) as r:
            s = r.structures(phases_class=cls, phase_class=cls._PHASE_CLASS)
        return s

    @classmethod
    def from_cif_string(cls, cif_string: str):
        s = None
        with Parsers('cif_str').reader() as r:
            s = r.structures(cif_string, phase_class=cls._PHASE_CLASS)
        return s
